#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio
import time
import aiohttp

import octobot_commons.logging as logging
import octobot_commons.constants as commons_constants
import octobot.constants as constants


async def get_current_octobots_stats():
    bot_metrics_url = f"{commons_constants.METRICS_URL}metrics/community/count/"
    return await _get_stats({
        "daily": f"{bot_metrics_url}0/0/-1",
        "monthly": f"{bot_metrics_url}0/-1/0",
        "all": f"{bot_metrics_url}0/0/0"
    })


async def _ensure_closed_sockets():
    # ugly workaround to prevent some of these current aiohttp version warnings:
    # "unclosed transport {'_pending_data': None, '_paused': False, '_sock': <socket.socket fd=-1,
    # family=AddressFamily.AF_INET, type=SocketKind.SOCK_STREAM, proto=6>,"
    await asyncio.sleep(0.01)


async def _get_stats(endpoint_by_key):
    logger = logging.get_logger("CommunityAnalysis")
    bots_stats = {}

    async def get_stats(url, stats_key):
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as resp:
                    if resp.status != 200:
                        logger.error(f"Error when getting community status : error code={resp.status}")
                    else:
                        json_resp = await resp.json()
                        if "total" in json_resp:
                            bots_stats[stats_key] = json_resp["total"]
                        else:
                            bots_stats[stats_key] = _format_top_elements(json_resp)
        except Exception as e:
            logger.exception(e, True, f"Error when getting community status : {e}")
    await asyncio.gather(
        *(
            get_stats(endpoint, key)
            for key, endpoint in endpoint_by_key.items()
        )
    )
    await _ensure_closed_sockets()
    return bots_stats


async def get_community_metrics():
    bot_count_metrics_url = f"{commons_constants.METRICS_URL}metrics/community/count/"
    bot_top_metrics_url = f"{commons_constants.METRICS_URL}metrics/community/top/"
    one_month_time = int(time.time() - 30 * commons_constants.DAYS_TO_SECONDS)
    simulated = "?traderType=simulated"
    real = "?traderType=real"
    return await _get_stats({
        "this_month": f"{bot_count_metrics_url}0/-1/0",
        "last_six_month": f"{bot_count_metrics_url}0/-6/0",
        "total_count": f"{bot_count_metrics_url}0/0/0",
        "top_simulated_exchanges": f"{bot_top_metrics_url}exchanges/{one_month_time}{simulated}",
        "top_real_exchanges": f"{bot_top_metrics_url}exchanges/{one_month_time}{real}",
        "top_simulated_pairs": f"{bot_top_metrics_url}pairs/{one_month_time}{simulated}",
        "top_real_pairs": f"{bot_top_metrics_url}pairs/{one_month_time}{real}",
        "top_simulated_eval_config": f"{bot_top_metrics_url}evaluation_configs/{one_month_time}{simulated}",
        "top_real_eval_config": f"{bot_top_metrics_url}evaluation_configs/{one_month_time}{real}",
    })


def can_read_metrics(config):
    return constants.IS_CLOUD_ENV or config.get_metrics_enabled()


def _format_top_elements(top_elements):
    return [
        {
            "name": element["name"],
            "count": element["count"],
            "rank": index,
        }
        for index, element in enumerate(top_elements)
    ]



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio

import octobot_commons.enums as common_enums
import octobot_commons.constants as common_constants

import octobot_trading.api as trading_api
import octobot_trading.octobot_channel_consumer as trading_channel_consumer

import octobot.channels as octobot_channel


class ExchangeProducer(octobot_channel.OctoBotChannelProducer):
    def __init__(self, channel, octobot, backtesting, ignore_config=False):
        super().__init__(channel)
        self.octobot = octobot
        self.ignore_config = ignore_config

        self.backtesting = backtesting
        self.exchange_manager_ids = []

        self.to_create_exchanges_count = 0
        self.created_all_exchanges = asyncio.Event()

    async def start(self):
        self.to_create_exchanges_count = 0
        self.created_all_exchanges.clear()
        for exchange_name in trading_api.get_enabled_exchanges_names(self.octobot.config):
            await self.create_exchange(exchange_name, self.backtesting)
            self.to_create_exchanges_count += 1

    def register_created_exchange_id(self, exchange_id):
        self.exchange_manager_ids.append(exchange_id)
        if len(self.exchange_manager_ids) == self.to_create_exchanges_count:
            self.created_all_exchanges.set()
            self.logger.debug(f"Exchange(s) created")

    async def stop(self):
        self.logger.debug("Stopping ...")
        for exchange_manager in trading_api.get_exchange_managers_from_exchange_ids(self.exchange_manager_ids):
            await trading_api.stop_exchange(exchange_manager)
        self.logger.debug("Stopped")

    async def create_exchange(self, exchange_name, backtesting):
        await self.send(bot_id=self.octobot.bot_id,
                        subject=common_enums.OctoBotChannelSubjects.CREATION.value,
                        action=trading_channel_consumer.OctoBotChannelTradingActions.EXCHANGE.value,
                        data={
                            trading_channel_consumer.OctoBotChannelTradingDataKeys.TENTACLES_SETUP_CONFIG.value:
                                self.octobot.tentacles_setup_config,
                            trading_channel_consumer.OctoBotChannelTradingDataKeys.MATRIX_ID.value:
                                self.octobot.evaluator_producer.matrix_id,
                            trading_channel_consumer.OctoBotChannelTradingDataKeys.BACKTESTING.value: backtesting,
                            trading_channel_consumer.OctoBotChannelTradingDataKeys.EXCHANGE_CONFIG.value:
                                self.octobot.config,
                            trading_channel_consumer.OctoBotChannelTradingDataKeys.EXCHANGE_NAME.value: exchange_name,
                        })



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

from octobot.community.feeds import abstract_feed
from octobot.community.feeds.abstract_feed import (
    AbstractFeed,
)
from octobot.community.feeds import community_ws_feed
from octobot.community.feeds.community_ws_feed import (
    CommunityWSFeed,
)
from octobot.community.feeds import community_mqtt_feed
from octobot.community.feeds.community_mqtt_feed import (
    CommunityMQTTFeed,
)
from octobot.community.feeds import community_supabase_feed
from octobot.community.feeds.community_supabase_feed import (
    CommunitySupabaseFeed,
)
from octobot.community.feeds import feed_factory
from octobot.community.feeds.feed_factory import (
    community_feed_factory,
)

__all__ = [
    "AbstractFeed",
    "CommunityWSFeed",
    "CommunityMQTTFeed",
    "CommunitySupabaseFeed",
    "community_feed_factory",
]



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import abc
import time

import octobot.automation.bases.automation_step as automation_step


class AbstractTriggerEvent(automation_step.AutomationStep, abc.ABC):
    def __init__(self):
        super(AbstractTriggerEvent, self).__init__()
        self.should_stop = False
        self.trigger_only_once = False
        self.max_trigger_frequency = 0
        self._last_trigger_time = 0

    async def stop(self):
        self.should_stop = True

    async def _get_next_event(self):
        raise NotImplementedError

    async def next_event(self):
        """
        Async generator, use as follows:
            async for event in self.next_event():
                # triggered when an event occurs
        """
        self._last_trigger_time = 0
        while not self.should_stop and not (self.trigger_only_once and self._last_trigger_time != 0):
            new_event = await self._get_next_event()
            trigger_time = time.time()
            if not self.max_trigger_frequency or (trigger_time - self._last_trigger_time > self.max_trigger_frequency):
                yield new_event
                self._last_trigger_time = time.time()



# pylint: disable=E0711, E0702
#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import octobot.community.models.community_supports as community_supports
import octobot.community.errors as errors
import octobot.community.supabase_backend.enums as backend_enums


class CommunityUserAccount:
    USER_DATA_CONTENT = "content"
    BOT_DEVICE = "device"
    BOT_URLS = "urls"
    BOT_DEPLOYMENT = "bot_deployment"
    BOT_DEPLOYMENT_TYPE = "type"
    METADATA = "metadata"
    FILLED_FORMS = "filledForms"
    NO_SELECTED_BOT_DESC = "No selected bot. Please select a bot to enable your community features."

    def __init__(self):
        self.gql_user_id = None
        self.bot_id = None
        self.gql_access_token = None
        self.supports = community_supports.CommunitySupports()

        self._profile_raw_data = None
        self._selected_bot_raw_data = None
        self._all_user_bots_raw_data = []

    def has_user_data(self):
        return self._profile_raw_data is not None

    def has_selected_bot_data(self):
        return self._selected_bot_raw_data is not None

    def get_email(self):
        return self._profile_raw_data[backend_enums.UserKeys.EMAIL.value]

    def get_user_id(self):
        return self._profile_raw_data[backend_enums.UserKeys.ID.value]

    def get_graph_token(self):
        raise NotImplemented
        return self._get_user_data_content()["graph_token"]

    def get_has_donated(self):
        return self._get_user_data_metadata().get("has_donated", False)

    def get_filled_forms_ids(self):
        return self._get_user_data_metadata().get(self.FILLED_FORMS, [])

    def get_all_user_bots_raw_data(self):
        return self._all_user_bots_raw_data

    def get_selected_bot_raw_data(self, raise_on_missing=False):
        if raise_on_missing:
            self._ensure_selected_bot_data()
        return self._selected_bot_raw_data

    def is_self_hosted(self, bot):
        return self._get_bot_deployment(bot).get(
            backend_enums.BotDeploymentKeys.TYPE.value, backend_enums.DeploymentTypes.SELF_HOSTED.value
        ) == backend_enums.DeploymentTypes.SELF_HOSTED.value

    def get_selected_bot_deployment_id(self):
        return self._get_bot_deployment(self._selected_bot_raw_data)[
            backend_enums.BotDeploymentKeys.ID.value
        ]

    def get_bot_deployment_status(self) -> (str, str):
        deployment = self._get_bot_deployment(self._selected_bot_raw_data)
        return (
            deployment[backend_enums.BotDeploymentKeys.STATUS.value],
            deployment[backend_enums.BotDeploymentKeys.DESIRED_STATUS.value]
        )

    def get_bot_deployment_url(self, deployment_url_data):
        return deployment_url_data[backend_enums.BotDeploymentURLKeys.URL.value]

    def get_selected_bot_device_uuid(self):
        raise NotImplemented
        try:
            return self.get_selected_bot_raw_data(raise_on_missing=True).get(self.BOT_DEVICE, {}).get("uuid", None)
        except AttributeError:
            raise errors.NoBotDeviceError("No device associated to the select bot")

    def get_selected_bot_device_name(self):
        raise NotImplemented
        try:
            return self.get_selected_bot_raw_data(raise_on_missing=True).get(self.BOT_DEVICE, {}).get("name", None)
        except AttributeError:
            raise errors.NoBotDeviceError("No device associated to the select bot")

    @staticmethod
    def get_bot_id(bot):
        return bot[backend_enums.BotKeys.ID.value]

    @staticmethod
    def get_bot_name_or_id(bot):
        return bot[backend_enums.BotKeys.NAME.value]

    def get_selected_bot_current_portfolio_id(self):
        return self._selected_bot_raw_data[backend_enums.BotKeys.CURRENT_PORTFOLIO_ID.value]

    def get_selected_bot_current_config_id(self):
        return self._selected_bot_raw_data[backend_enums.BotKeys.CURRENT_CONFIG_ID.value]

    def set_profile_raw_data(self, profile_raw_data):
        self._profile_raw_data = profile_raw_data

    def set_selected_bot_raw_data(self, selected_bot_raw_data):
        self._selected_bot_raw_data = selected_bot_raw_data

    def set_selected_bot_device_raw_data(self, bot_raw_data_with_device):
        device_data = bot_raw_data_with_device[self.BOT_DEVICE]
        if device_data is None:
            raise errors.NoBotDeviceError("No device associated to the select bot")
        self._selected_bot_raw_data[self.BOT_DEVICE] = device_data

    def set_all_user_bots_raw_data(self, all_user_bots_raw_data):
        self._all_user_bots_raw_data = all_user_bots_raw_data

    def _get_user_data_content(self):
        return self._profile_raw_data[self.USER_DATA_CONTENT]

    def _get_user_data_metadata(self):
        return self._profile_raw_data.get(backend_enums.UserKeys.USER_METADATA.value, {})

    def _get_bot_deployment(self, bot):
        if bot is None:
            raise errors.BotError(self.NO_SELECTED_BOT_DESC)
        return bot.get(self.BOT_DEPLOYMENT) or {}

    def _ensure_selected_bot_data(self):
        if self._selected_bot_raw_data is None:
            raise errors.BotError(self.NO_SELECTED_BOT_DESC)

    def ensure_selected_bot_id(self):
        if self.bot_id is None:
            raise errors.BotError("No selected bot")

    def get_support_role(self):
        try:
            if self.get_has_donated():
                return community_supports.CommunitySupports.OCTOBOT_DONOR_ROLE
        except KeyError:
            pass
        return community_supports.CommunitySupports.DEFAULT_SUPPORT_ROLE

    def flush_bot_details(self):
        self.bot_id = None
        self._selected_bot_raw_data = None
        self._all_user_bots_raw_data = []

    def flush(self):
        self.gql_user_id = None
        self.gql_access_token = None
        self.supports = community_supports.CommunitySupports()
        self._profile_raw_data = None
        self.flush_bot_details()



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio
import pytest

import tentacles

import octobot_commons.asyncio_tools as asyncio_tools
from octobot_trading.api.exchange import get_exchange_manager_from_exchange_id
from octobot_trading.api.orders import get_open_orders
from octobot_trading.api.profitability import get_profitability_stats
from octobot.api.backtesting import get_independent_backtesting_exchange_manager_ids, stop_independent_backtesting, \
    check_independent_backtesting_remaining_objects
from octobot.backtesting.abstract_backtesting_test import DATA_FILES
from octobot_trading.api.trades import get_trade_history
from tests.test_utils.bot_management import run_independent_backtesting
from tests.test_utils.logging import activate_tentacles_loading_logger
import octobot_tentacles_manager.api as tentacles_manager_api

activate_tentacles_loading_logger()

BACKTESTING_SYMBOLS = ["ICX/BTC", "VEN/BTC", "XRB/BTC", "2020ADA/BTC", "2020ADA/USDT", "BTC/USDT"]

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.fixture
def event_loop():
    loop = asyncio.new_event_loop()
    tentacles_manager_api.reload_tentacle_info()
    # use ErrorContainer to catch otherwise hidden exceptions occurring in async scheduled tasks
    error_container = asyncio_tools.ErrorContainer()
    loop.set_exception_handler(error_container.exception_handler)
    yield loop
    # will fail if exceptions have been silently raised
    loop.run_until_complete(error_container.check())
    loop.close()


async def test_single_data_file_backtesting():
    await _check_double_backtesting(
        run_independent_backtesting([DATA_FILES[BACKTESTING_SYMBOLS[0]]], enable_storage=True),
        run_independent_backtesting([DATA_FILES[BACKTESTING_SYMBOLS[0]]], enable_storage=True)
    )


async def test_single_data_file_no_logs_backtesting():
    await _check_double_backtesting(
        run_independent_backtesting([DATA_FILES[BACKTESTING_SYMBOLS[0]]], use_loggers=False),
        run_independent_backtesting([DATA_FILES[BACKTESTING_SYMBOLS[0]]], use_loggers=False)
    )


async def test_single_data_file_mixed_logs_backtesting():
    await _check_double_backtesting(
        run_independent_backtesting([DATA_FILES[BACKTESTING_SYMBOLS[0]]], use_loggers=False),
        run_independent_backtesting([DATA_FILES[BACKTESTING_SYMBOLS[0]]], use_loggers=True)
    )


async def test_double_synchronized_data_file_backtesting():
    await _check_double_backtesting(
        run_independent_backtesting([DATA_FILES[BACKTESTING_SYMBOLS[0]],
                                     DATA_FILES[BACKTESTING_SYMBOLS[1]]],
                                    timeout=40),
        run_independent_backtesting([DATA_FILES[BACKTESTING_SYMBOLS[0]],
                                     DATA_FILES[BACKTESTING_SYMBOLS[1]]],
                                    timeout=40)
    )


async def test_double_partially_synchronized_data_file_backtesting_common_only():
    await _check_double_backtesting(
        run_independent_backtesting([DATA_FILES[BACKTESTING_SYMBOLS[3]],
                                     DATA_FILES[BACKTESTING_SYMBOLS[4]]],
                                    timeout=40),
        run_independent_backtesting([DATA_FILES[BACKTESTING_SYMBOLS[3]],
                                     DATA_FILES[BACKTESTING_SYMBOLS[4]]],
                                    timeout=40)
    )


async def test_double_partially_synchronized_data_file_backtesting_all_data_files_range():
    await _check_double_backtesting(
        run_independent_backtesting([DATA_FILES[BACKTESTING_SYMBOLS[3]],
                                     DATA_FILES[BACKTESTING_SYMBOLS[4]]],
                                    run_on_common_part_only=False,
                                    timeout=40),
        run_independent_backtesting([DATA_FILES[BACKTESTING_SYMBOLS[3]],
                                     DATA_FILES[BACKTESTING_SYMBOLS[4]]],
                                    run_on_common_part_only=False,
                                    timeout=40)
    )


async def test_double_data_file_3_months_gap_backtesting_all_data_files_range():
    await _check_double_backtesting(
        run_independent_backtesting([DATA_FILES[BACKTESTING_SYMBOLS[5]],
                                     DATA_FILES[BACKTESTING_SYMBOLS[0]]],
                                    run_on_common_part_only=False,
                                    timeout=80),
        run_independent_backtesting([DATA_FILES[BACKTESTING_SYMBOLS[5]],
                                     DATA_FILES[BACKTESTING_SYMBOLS[0]]],
                                    run_on_common_part_only=False,
                                    timeout=80)
    )


async def _check_double_backtesting(backtesting_coro1, backtesting_coro2):
    backtesting1 = None
    backtesting2 = None
    try:
        backtesting1 = await backtesting_coro1
        backtesting2 = await backtesting_coro2
    finally:
        await _check_backtesting_results(backtesting1, backtesting2)


async def _check_backtesting_results(backtesting1, backtesting2):
    if backtesting1 is not None and backtesting2 is not None:
        exchange_manager_1 = get_exchange_manager_from_exchange_id(
            get_independent_backtesting_exchange_manager_ids(backtesting1)[0])
        exchange_manager_2 = get_exchange_manager_from_exchange_id(
            get_independent_backtesting_exchange_manager_ids(backtesting2)[0])

        _, previous_profitability, previous_market_profitability, _, _ = get_profitability_stats(exchange_manager_1)
        _, current_profitability, current_market_profitability, _, _ = get_profitability_stats(exchange_manager_2)

        trades = get_trade_history(exchange_manager_1)
        open_orders = get_open_orders(exchange_manager_1)
        # ensure at least one order is either open or got filled
        assert trades + open_orders

        trades = open_orders = exchange_manager_1 = exchange_manager_2 = None  # prevent memory leak

        # ensure no randomness in backtesting
        assert previous_profitability == current_profitability
        assert current_profitability != 0
        assert previous_profitability != 0
        assert previous_market_profitability == current_market_profitability

    backtestings = tuple(b for b in (backtesting1, backtesting2) if b is not None)
    for backtesting in backtestings:
        await stop_independent_backtesting(backtesting, memory_check=True, should_raise=True)
        await asyncio.wait_for(backtesting.post_backtesting_task, 5)
    for backtesting in backtestings:
        asyncio.get_event_loop().call_soon(check_independent_backtesting_remaining_objects, backtesting)



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import octobot_tentacles_manager.api as tentacles_manager_api
import octobot.constants as constants
import octobot_commons.databases as databases
import octobot_commons.logging as logging
import octobot_commons.errors as commons_errors
import octobot.databases_util as databases_util


class Initializer:
    """Initializer class:
    - Initialize services, constants and tools
    """

    def __init__(self, octobot):
        self.octobot = octobot

    async def create(self, init_bot_storage):
        # initialize tentacle configuration
        tentacles_config_path = self.octobot.get_startup_config(constants.CONFIG_KEY, dict_only=False).\
            get_tentacles_config_path()
        self.octobot.tentacles_setup_config = tentacles_manager_api.get_tentacles_setup_config(tentacles_config_path)

        if init_bot_storage:
            try:
                # init bot storage
                await databases.init_bot_storage(
                    self.octobot.bot_id,
                    databases_util.get_run_databases_identifier(
                        self.octobot.config,
                        self.octobot.tentacles_setup_config
                    ),
                    True
                )
            except commons_errors.ConfigTradingError as err:
                # already logged as error, don't display it twice
                logging.get_logger(self.__class__.__name__).warning(f"Error when initializing bot storage: {err}")
            except Exception as err:
                logging.get_logger(self.__class__.__name__).exception(
                    err, True, f"Error when initializing bot storage: {err}"
                )

        # create OctoBot channel
        await self.octobot.global_consumer.initialize()



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import time

import octobot.community.errors_upload.errors_uploader as errors_uploader
import octobot.community.errors_upload.error_model as error_model
import octobot_commons.logging as logging
import octobot.constants as constants


class _UploadWrapper:
    def __init__(self, upload_url, config):
        self._config = config
        self._metrics_id = self._get_metrics_id()
        self._uploader = errors_uploader.ErrorsUploader(
            upload_url
        )

    def upload_if_necessary(self, exception, error_message):
        if constants.UPLOAD_ERRORS and (constants.IS_CLOUD_ENV or self._config.get_metrics_enabled()):
            self._uploader.schedule_error_upload(
                error_model.Error(
                    exception, error_message, time.time(), self._metrics_id
                )
            )

    def _get_metrics_id(self):
        try:
            return self._config.get_metrics_id()
        except KeyError:
            return constants.DEFAULT_METRICS_ID


def register_error_uploader(upload_url, config):
    upload_wrapper = _UploadWrapper(upload_url, config)
    logging.BotLogger.register_error_callback(upload_wrapper.upload_if_necessary)



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import octobot.community.models.community_donation as community_donation
import octobot_commons.support as support


class CommunitySupports(support.Support):
    DEFAULT_SUPPORT_ROLE = "default"
    OCTOBOT_DONOR_ROLE = "donor"

    def __init__(self, support_role: str = None, donations: list = None):
        self.support_role = support_role or CommunitySupports.DEFAULT_SUPPORT_ROLE
        self.donations = donations or []

    def is_supporting(self) -> bool:
        return self.support_role != self.DEFAULT_SUPPORT_ROLE or self.is_donor()

    def is_donor(self) -> bool:
        return self.support_role == self.OCTOBOT_DONOR_ROLE or bool(self.donations)

    @staticmethod
    def from_community_dict(data):
        return CommunitySupports(
            data["data"]["attributes"].get("support_role", CommunitySupports.DEFAULT_SUPPORT_ROLE),
            [community_donation.CommunityDonation.from_community_dict(donation_data)
             for donation_data in data.get("included", [])]
        )



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import contextlib
import mock
import pytest
import pytest_asyncio
import sentry_sdk

import octobot.community.errors_upload.sentry_aiohttp_transport
import octobot_commons.logging
import octobot_commons.asyncio_tools


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


SENTRY_CONFIG = {
    "enabled": False
}


@pytest_asyncio.fixture
async def sentry_init():
    handle = None
    SENTRY_CONFIG["enabled"] = True
    try:
        def _before_send_callback(event: dict, hint: dict):
            # disable after test to avoid spamming in other tests in case client.close does not work
            if SENTRY_CONFIG["enabled"]:
                return event
            return None

        handle = sentry_sdk.init(
            dsn="https://9ad6367a0c2a4b7d8a3014ab87a2d96e@plop.com/1",

            # Set traces_sample_rate to 1.0 to capture 100%
            # of transactions for performance monitoring.
            # We recommend adjusting this value in production.
            traces_sample_rate=1.0,

            before_send=_before_send_callback,
            transport=octobot.community.errors_upload.sentry_aiohttp_transport.SentryAiohttpTransport,
        )
        handle._client.transport._worker.start()
        yield handle._client.transport
    finally:
        SENTRY_CONFIG["enabled"] = False
        if handle and hasattr(handle._client.transport, "async_kill"):
            await handle._client.transport.async_kill()
        client = sentry_sdk.Hub.current.client
        if client is not None:
            client.close(timeout=0)


def _mocked_context(return_value):
    call_mock = mock.AsyncMock()

    @contextlib.asynccontextmanager
    async def _inner(*args, **kwargs):
        await call_mock(*args, **kwargs)
        yield return_value
    return _inner, call_mock


def _mocked_resp():
    return mock.Mock(
        status=200,
        headers={},
        text=mock.AsyncMock(return_value="plop")
    )


async def test_upload_error(sentry_init):
    transport = sentry_init
    patched_post, post_mock = _mocked_context(_mocked_resp())
    logger = octobot_commons.logging.get_logger(__name__)
    with mock.patch.object(transport._worker.session, "post", patched_post):
        # test with error
        logger.error("error message hello")
        post_mock.assert_not_called()
        await octobot_commons.asyncio_tools.wait_asyncio_next_cycle()
        # if post_mock is called, then the whole transport works as the aiohttp session post is called by sentry
        assert len(post_mock.mock_calls) == 1
        assert post_mock.mock_calls[0].args == ('https://plop.com/api/1/envelope/',)
        assert list(post_mock.mock_calls[0].kwargs) == ["data", "headers"]

        post_mock.reset_mock()
        # test with exception
        try:
            1/0
        except ZeroDivisionError as err:
            logger.exception(err, True, f"demo exception {err}")
        await octobot_commons.asyncio_tools.wait_asyncio_next_cycle()
        # if post_mock is called, then the whole transport works as the aiohttp session post is called by sentry
        assert len(post_mock.mock_calls) == 2   # called twice to account for error and exception log
        for call in post_mock.mock_calls:
            assert call.args == ('https://plop.com/api/1/envelope/',)
            assert list(call.kwargs) == ["data", "headers"]

        post_mock.reset_mock()
        try:
            1/0
        except ZeroDivisionError as err:
            logger.exception(err, False, f"demo exception {err}")
        await octobot_commons.asyncio_tools.wait_asyncio_next_cycle()
        # if post_mock is called, then the whole transport works as the aiohttp session post is called by sentry
        assert len(post_mock.mock_calls) == 1   # called for exception log only
        for call in post_mock.mock_calls:
            assert call.args == ('https://plop.com/api/1/envelope/',)
            assert list(call.kwargs) == ["data", "headers"]



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


class ScoredRunResult:
    def __init__(self, full_result, optimizer_run_data):
        self.full_result = full_result
        self.optimizer_run_data = optimizer_run_data
        self.values = {}
        self.score = 0
        self.total_weight = 0

    def compute_score(self, relevant_scoring_parameters):
        self.score = 0
        try:
            self.score = sum([
                self._compute_score(scoring_parameter)
                for scoring_parameter in relevant_scoring_parameters
            ]) / self.total_weight
        except ZeroDivisionError:
            self.score = 0

    def _compute_score(self, fitness_parameter):
        try:
            self.values[fitness_parameter.name] = self.full_result[fitness_parameter.name]
            score = fitness_parameter.get_normalized_value(self.values[fitness_parameter.name])
            self.total_weight += fitness_parameter.weight
            return score
        except KeyError:
            return 0

    def __repr__(self):
        return f"[{self.__class__.__name__}] score: {self.score}, total_weight: {self.total_weight}"

    def result_str(self):
        # todo move constants outside of StrategyDesignOptimizer
        import octobot.strategy_optimizer.strategy_design_optimizer as strategy_design_optimizer
        user_inputs = {
            ui[strategy_design_optimizer.StrategyDesignOptimizer.CONFIG_USER_INPUT]:
                ui[strategy_design_optimizer.StrategyDesignOptimizer.CONFIG_VALUE]
            for ui in self.optimizer_run_data
        }
        return f"fitness score: {self.score} {self.values} from {user_inputs}"



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import pytest
import mock
import contextlib
import asyncio
import aiohttp

import octobot.community as community
from tests.unit_tests.community.errors_upload import error_uploader, basic_error, exception_error


def test_schedule_error_upload(error_uploader, basic_error):
    with mock.patch.object(error_uploader, "_add_error", mock.Mock()) as _add_error_mock, \
         mock.patch.object(error_uploader, "_ensure_upload_task", mock.Mock()) as _ensure_upload_task_mock:
        error_uploader.schedule_error_upload(basic_error)
        _add_error_mock.assert_called_once_with(basic_error)
        _ensure_upload_task_mock.assert_called_once()


def test_add_error(error_uploader, basic_error):
    with mock.patch.object(basic_error, "merge_equivalent", mock.Mock()) as merge_equivalent_mock:
        with mock.patch.object(basic_error, "is_equivalent", mock.Mock(return_value=False)) as is_equivalent_mock:
            error_uploader._add_error(basic_error)
            merge_equivalent_mock.assert_not_called()
            is_equivalent_mock.assert_not_called()
            assert error_uploader._to_upload_errors == [basic_error]

            error_uploader._add_error(basic_error)
            merge_equivalent_mock.assert_not_called()
            is_equivalent_mock.assert_called_once_with(basic_error)
            assert error_uploader._to_upload_errors == [basic_error, basic_error]

        with mock.patch.object(basic_error, "is_equivalent", mock.Mock(return_value=True)) as is_equivalent_mock:
            error_uploader._add_error(basic_error)
            merge_equivalent_mock.assert_called_once_with(basic_error)
            is_equivalent_mock.assert_called_once_with(basic_error)
            assert error_uploader._to_upload_errors == [basic_error, basic_error]


def test_ensure_upload_task(error_uploader):
    error_uploader.logger = mock.Mock()
    with mock.patch.object(error_uploader.logger, "exception", mock.Mock()) as exception_mock:
        with mock.patch.object(error_uploader, "_schedule_upload", mock.Mock()) as _schedule_upload_mock:
            with mock.patch.object(error_uploader, "_ensure_event_loop", mock.Mock(return_value=False)) \
                    as _ensure_event_loop_mock:
                error_uploader._ensure_upload_task()
                _ensure_event_loop_mock.assert_called_once()
                _schedule_upload_mock.assert_not_called()
                exception_mock.assert_not_called()
            with mock.patch.object(error_uploader, "_ensure_event_loop", mock.Mock(return_value=True)) \
                    as _ensure_event_loop_mock:
                error_uploader._ensure_upload_task()
                _ensure_event_loop_mock.assert_called_once()
                _schedule_upload_mock.assert_called_once()
                exception_mock.assert_not_called()
                _ensure_event_loop_mock.reset_mock()
                _schedule_upload_mock.reset_mock()

                error_uploader._upload_task = mock.Mock()
                with mock.patch.object(error_uploader._upload_task, "done", mock.Mock(return_value=False)) \
                        as _upload_task_mock:
                    error_uploader._ensure_upload_task()
                    _ensure_event_loop_mock.assert_called_once()
                    _upload_task_mock.assert_called_once()
                    _schedule_upload_mock.assert_not_called()
                    exception_mock.assert_not_called()
                    _ensure_event_loop_mock.reset_mock()
                with mock.patch.object(error_uploader._upload_task, "done", mock.Mock(return_value=True)) \
                        as _upload_task_mock:
                    error_uploader._ensure_upload_task()
                    _ensure_event_loop_mock.assert_called_once()
                    _upload_task_mock.assert_called_once()
                    _schedule_upload_mock.assert_called_once()
                    exception_mock.assert_not_called()
                    _ensure_event_loop_mock.reset_mock()
                with mock.patch.object(error_uploader._upload_task, "done", mock.Mock(side_effect=RuntimeError())) \
                        as _upload_task_mock:
                    error_uploader._ensure_upload_task()
                    _ensure_event_loop_mock.assert_called_once()
                    _upload_task_mock.assert_called_once()
                    _schedule_upload_mock.assert_called_once()
                    exception_mock.assert_called_once()


@pytest.mark.asyncio
async def test_upload_errors(error_uploader, basic_error, exception_error):
    session = mock.Mock()
    resp = mock.Mock()
    resp.status = 400
    resp.text = mock.AsyncMock(return_value="text")
    post_mock = mock.AsyncMock(return_value=resp)

    @contextlib.asynccontextmanager
    async def post(*args, **kwargs):
        yield await post_mock(*args, **kwargs)
    session.post = post
    errors = [basic_error, exception_error]
    error_uploader.logger = mock.Mock()
    with mock.patch.object(error_uploader.logger, "error", mock.Mock(return_value=None)) as error_mock, \
         mock.patch.object(error_uploader, "_get_formatted_errors", mock.Mock(return_value=None)) \
         as _get_formatted_errors_mock:
        await error_uploader._upload_errors(session, errors)
        post_mock.assert_called_once_with(error_uploader.upload_url, json=None)
        error_mock.assert_called_once()
        post_mock.reset_mock()
        error_mock.reset_mock()

        resp.status = 200
        await error_uploader._upload_errors(session, errors)
        post_mock.assert_called_once_with(error_uploader.upload_url, json=None)
        error_mock.assert_not_called()


def test_get_formatted_errors(error_uploader, basic_error, exception_error):
    with mock.patch.object(community.Error, "to_dict", mock.Mock(return_value={})) as to_dict_mock:
        assert error_uploader._get_formatted_errors([basic_error, exception_error]) == [{}, {}]
        assert to_dict_mock.call_count == 2


def test_schedule_upload(error_uploader):
    error_uploader.loop = mock.Mock()
    error_uploader.loop.create_task = mock.Mock(return_value="return")
    with mock.patch.object(error_uploader, "_upload_soon", mock.Mock(return_value="return")) as _upload_soon_mock:
        error_uploader._schedule_upload()
        _upload_soon_mock.assert_called_once()
        error_uploader.loop.create_task.assert_called_once_with("return")
        assert error_uploader._upload_task == "return"


@pytest.mark.asyncio
async def test_upload_soon(error_uploader, basic_error):
    _session_mock = mock.AsyncMock(return_value="plop")

    @contextlib.asynccontextmanager
    async def session_mock(*args, **kwargs):
        yield await _session_mock(*args, **kwargs)
    with mock.patch.object(asyncio, "sleep", mock.AsyncMock()) as sleep_mock, \
         mock.patch.object(aiohttp, "ClientSession", session_mock), \
         mock.patch.object(error_uploader, "_schedule_upload", mock.Mock()) as _schedule_upload_mock:
        with mock.patch.object(error_uploader, "_upload_errors", mock.AsyncMock()) as _upload_errors_mock:

            # no error to upload
            await error_uploader._upload_soon()
            sleep_mock.assert_called_once_with(error_uploader.upload_delay)
            _session_mock.assert_not_called()
            _upload_errors_mock.assert_not_called()
            _schedule_upload_mock.assert_not_called()
            sleep_mock.reset_mock()

            # errors to upload
            error_uploader._to_upload_errors = [basic_error]
            await error_uploader._upload_soon()
            assert error_uploader._to_upload_errors == []
            sleep_mock.assert_called_once_with(error_uploader.upload_delay)
            _session_mock.assert_called_once()
            _upload_errors_mock.assert_called_once_with("plop", [basic_error])
            _schedule_upload_mock.assert_not_called()
            sleep_mock.reset_mock()
            _session_mock.reset_mock()
            _upload_errors_mock.reset_mock()

        async def upload_errors_adding_errors(*_):
            error_uploader._to_upload_errors = [basic_error]
        with mock.patch.object(error_uploader, "_upload_errors", upload_errors_adding_errors):
            # errors to upload
            error_uploader._to_upload_errors = [basic_error]
            await error_uploader._upload_soon()
            sleep_mock.assert_called_once_with(error_uploader.upload_delay)
            _session_mock.assert_called_once()
            # reschedule call
            _schedule_upload_mock.assert_called_once()
            sleep_mock.reset_mock()
            _session_mock.reset_mock()


def test_ensure_event_loop(error_uploader):
    error_uploader.loop = mock.Mock()
    with mock.patch.object(asyncio, "get_event_loop", mock.Mock(return_value="loop")) as get_event_loop_mock:
        with mock.patch.object(error_uploader.loop, "is_running", mock.Mock(return_value=True)) as is_running_mock:
            error_uploader._ensure_event_loop()
            is_running_mock.assert_called_once()
            get_event_loop_mock.assert_not_called()
        with mock.patch.object(error_uploader.loop, "is_running", mock.Mock(return_value=False)) as is_running_mock:
            assert error_uploader._ensure_event_loop()
            is_running_mock.assert_called_once()
            get_event_loop_mock.assert_called_once()
            assert error_uploader.loop == "loop"
    error_uploader.loop = None
    with mock.patch.object(asyncio, "get_event_loop", mock.Mock(side_effect=RuntimeError())) as get_event_loop_mock:
        assert error_uploader._ensure_event_loop() is False
        get_event_loop_mock.assert_called_once()



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import os

import mock
import builtins

import tests.test_utils.config as test_utils_config
import octobot_commons.tests.test_config as test_config
import tentacles.Evaluator.Strategies as tentacles_strategies
import octobot.strategy_optimizer as strategy_optimizer


class StrategyTestSuiteMock(mock.Mock):
    def __init__(self, *args, **kwargs):
        super(StrategyTestSuiteMock, self).__init__(*args, **kwargs)
        self.run_test_suite = mock.AsyncMock()
        self.initialize_with_strategy = mock.Mock()

    def get_test_suite_result(self, *_, **__):
        strategy_name = tentacles_strategies.SimpleStrategyEvaluator.get_name()
        return strategy_optimizer.TestSuiteResult([(1, 0), (0, -1)], [1, 2], 1, ["tf1", "tf2"], [strategy_name], strategy_name)


def test_find_optimal_configuration():
    with mock.patch.object(strategy_optimizer, "StrategyTestSuite", StrategyTestSuiteMock()) as test_suite_mock, \
         mock.patch.object(builtins, "print", mock.Mock()) as print_mock:
        strategy_name = tentacles_strategies.SimpleStrategyEvaluator.get_name()
        optimizer = strategy_optimizer.StrategyOptimizer(test_config.load_test_config(),
                                                         test_utils_config.load_test_tentacles_config(),
                                                         strategy_name)
        optimizer.find_optimal_configuration()
        if os.getenv('CYTHON_IGNORE'):
            return
        assert optimizer.total_nb_runs == 21
        assert test_suite_mock.call_count == optimizer.total_nb_runs
        assert print_mock.call_count == optimizer.total_nb_runs * 2
        # check each call has been different
        # iterate over each second print call to check run config (each strategy optimizer run prints twice)
        for call in print_mock.call_args_list[::2]:
            assert len([c for c in print_mock.call_args_list if c == call]) == 1



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import typing
import asyncio
import aiohttp
import io
import gzip

try:
    import sentry_sdk
    import sentry_sdk.consts
    import sentry_sdk.utils
    import sentry_sdk.envelope
    import sentry_sdk.transport
    import sentry_sdk._types

    class SentryAiohttpTransport(sentry_sdk.HttpTransport):
        def __init__(
            self, options: typing.Dict[str, typing.Any]
        ):
            super().__init__(options)
            # use custom async worker instead of default sentry thread worker
            # does not support proxies, at least for now
            self._worker = AiohttpWorker(queue_size=options["transport_queue_size"])

        async def _async_send_request(
            self,
            body,  # type: bytes
            headers,  # type: typing.Dict[str, str]
            endpoint_type="store",  # type: sentry_sdk._types.EndpointType
            envelope=None,  # type: typing.Optional[sentry_sdk.envelope.Envelope]
        ) -> None:

            def record_loss(reason: str) -> None:
                if envelope is None:
                    self.record_lost_event(reason, data_category="error")
                else:
                    for item in envelope.items:
                        self.record_lost_event(reason, item=item)

            headers.update(
                {
                    "User-Agent": str(self._auth.client),
                    "X-Sentry-Auth": str(self._auth.to_header()),
                }
            )
            try:
                async with self._worker.session.post(
                    str(self._auth.get_api_url(endpoint_type)),
                    data=body,
                    headers=headers,
                ) as response:
                    self._update_rate_limits(response)

                    if response.status == 429:
                        # if we hit a 429.  Something was rate limited but we already
                        # acted on this in `self._update_rate_limits`.  Note that we
                        # do not want to record event loss here as we will have recorded
                        # an outcome in relay already.
                        self.on_dropped_event("status_429")
                        pass

                    elif response.status >= 300 or response.status < 200:
                        sentry_sdk.utils.logger.error(
                            "Unexpected status code: %s (body: %s)",
                            response.status,
                            await response.text(),
                        )
                        self.on_dropped_event("status_{}".format(response.status))
                        record_loss("network_error")
            except Exception:
                self.on_dropped_event("network")
                record_loss("network_error")
                raise

        async def _async_send_event(
            self, event  # type: sentry_sdk._types.Event
        ) -> None:

            if self._check_disabled("error"):
                self.on_dropped_event("self_rate_limits")
                self.record_lost_event("ratelimit_backoff", data_category="error")
                return None

            body = io.BytesIO()
            with gzip.GzipFile(fileobj=body, mode="w") as f:
                f.write(sentry_sdk.utils.json_dumps(event))

            assert self.parsed_dsn is not None
            sentry_sdk.utils.logger.debug(
                "Sending event, type:%s level:%s event_id:%s project:%s host:%s"
                % (
                    event.get("type") or "null",
                    event.get("level") or "null",
                    event.get("event_id") or "null",
                    self.parsed_dsn.project_id,
                    self.parsed_dsn.host,
                )
            )
            await self._async_send_request(
                body.getvalue(),
                headers={"Content-Type": "application/json", "Content-Encoding": "gzip"},
            )
            return None

        async def _async_send_envelope(
            self, envelope: sentry_sdk.envelope.Envelope
        ) -> None:

            # remove all items from the envelope which are over quota
            new_items = []
            for item in envelope.items:
                if self._check_disabled(item.data_category):
                    if item.data_category in ("transaction", "error", "default"):
                        self.on_dropped_event("self_rate_limits")
                    self.record_lost_event("ratelimit_backoff", item=item)
                else:
                    new_items.append(item)

            # Since we're modifying the envelope here make a copy so that others
            # that hold references do not see their envelope modified.
            envelope = sentry_sdk.envelope.Envelope(headers=envelope.headers, items=new_items)

            if not envelope.items:
                return None

            # since we're already in the business of sending out an envelope here
            # check if we have one pending for the stats session envelopes so we
            # can attach it to this enveloped scheduled for sending.  This will
            # currently typically attach the client report to the most recent
            # session update.
            client_report_item = self._fetch_pending_client_report(interval=30)
            if client_report_item is not None:
                envelope.items.append(client_report_item)

            body = io.BytesIO()
            with gzip.GzipFile(fileobj=body, mode="w") as f:
                envelope.serialize_into(f)

            assert self.parsed_dsn is not None
            sentry_sdk.utils.logger.debug(
                "Sending envelope [%s] project:%s host:%s",
                envelope.description,
                self.parsed_dsn.project_id,
                self.parsed_dsn.host,
            )

            await self._async_send_request(
                body.getvalue(),
                headers={
                    "Content-Type": "application/x-sentry-envelope",
                    "Content-Encoding": "gzip",
                },
                endpoint_type="envelope",
                envelope=envelope,
            )
            return None

        def capture_event(
            self, event  # type: sentry_sdk._types.Event
        ) -> None:
            hub = self.hub_cls.current

            async def send_event_wrapper() -> None:
                with hub:    # pylint: disable=not-context-manager
                    with sentry_sdk.utils.capture_internal_exceptions():
                        await self._async_send_event(event)
                        self._flush_client_reports()

            if not self._worker.submit(send_event_wrapper):
                self.on_dropped_event("full_queue")
                self.record_lost_event("queue_overflow", data_category="error")

        def capture_envelope(
            self, envelope: sentry_sdk.envelope.Envelope
        ) -> None:
            hub = self.hub_cls.current

            async def send_envelope_wrapper() -> None:
                with hub:    # pylint: disable=not-context-manager
                    with sentry_sdk.utils.capture_internal_exceptions():
                        await self._async_send_envelope(envelope)
                        self._flush_client_reports()

            if not self._worker.submit(send_envelope_wrapper):
                self.on_dropped_event("full_queue")
                for item in envelope.items:
                    self.record_lost_event("queue_overflow", item=item)

        async def async_kill(self):
            await self._worker.async_kill()

    class AiohttpWorker:
        def __init__(self, queue_size=sentry_sdk.consts.DEFAULT_QUEUE_SIZE):
            self.session = None
            self.call_tasks = []
            self._queue_size = queue_size
            self._kill_task = None
            self._stopped = False

        @property
        def is_alive(self) -> bool:
            if self.session is None or self.session.closed:
                return False
            # session is not closed
            # no pending kill task
            if self._kill_task is None or self._kill_task.done():
                return not self.session.closed
            # pending kill task, will stop
            return False

        def _ensure_session(self) -> None:
            if not self.is_alive:
                self.start()

        def start(self) -> None:
            if self._kill_task and not self._kill_task.done():
                self._kill_task.cancel()
            self.session = aiohttp.ClientSession()

        def kill(self) -> None:
            for task in self.call_tasks:
                if not task.done():
                    task.cancel()
            self.call_tasks = []
            if self.is_alive:
                self._kill_task = asyncio.create_task(self.session.close())

        async def async_kill(self):
            self._stopped = type
            self.kill()
            if self._kill_task and not self._kill_task.done():
                await self._kill_task

        def full(self) -> bool:
            return len(self.call_tasks) > self._queue_size

        def flush(self, timeout: float, callback=None) -> None:
            sentry_sdk.utils.logger.debug("background worker got flush request")
            sentry_sdk.utils.logger.debug("background worker flush ignored")

        async def _async_call(self, callback):
            try:
                await callback()
            finally:
                if asyncio.current_task() in self.call_tasks:
                    self.call_tasks.remove(asyncio.current_task())

        def _schedule_async_send(self, callback):
            self.call_tasks.append(asyncio.create_task(self._async_call(callback)))

        def submit(self, callback) -> bool:
            if self._stopped:
                return False
            self._ensure_session()
            if self.full():
                return False
            self._schedule_async_send(callback)
            return True

except ImportError:
    # do not force sentry_sdk requirement
    pass



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_commons.enums as commons_enums
import octobot.enums as enums
import octobot.constants as constants
import octobot.strategy_optimizer.fitness_parameter as fitness_parameter
import octobot.strategy_optimizer.optimizer_filter as optimizer_filter
import octobot.strategy_optimizer.optimizer_constraint as optimizer_constraint


class OptimizerSettings:
    def __init__(self, settings_dict=None):
        if settings_dict is None:
            settings_dict = {}
        # generic
        self.optimizer_config = settings_dict.get(enums.OptimizerConfig.OPTIMIZER_CONFIG.value, None)
        self.randomly_chose_runs = settings_dict.get(enums.OptimizerConfig.RANDOMLY_CHOSE_RUNS.value,
                                                     constants.OPTIMIZER_DEFAULT_RANDOMLY_CHOSE_RUNS)
        self.data_files = settings_dict.get(enums.OptimizerConfig.DATA_FILES.value)
        self.start_timestamp = settings_dict.get(enums.OptimizerConfig.START_TIMESTAMP.value, None)
        self.end_timestamp = settings_dict.get(enums.OptimizerConfig.END_TIMESTAMP.value, None)
        self.required_idle_cores = int(settings_dict.get(enums.OptimizerConfig.IDLE_CORES.value,
                                                         constants.OPTIMIZER_DEFAULT_REQUIRED_IDLE_CORES))
        self.notify_when_complete = settings_dict.get(enums.OptimizerConfig.NOTIFY_WHEN_COMPLETE.value,
                                                      constants.OPTIMIZER_DEFAULT_NOTIFY_WHEN_COMPLETE)
        self.optimizer_mode = settings_dict.get(enums.OptimizerConfig.MODE.value,
                                                enums.OptimizerModes.NORMAL.value)
        optimizer_id = settings_dict.get(enums.OptimizerConfig.OPTIMIZER_ID.value, 1)
        self.optimizer_id = optimizer_id if optimizer_id is None else int(optimizer_id)
        self.optimizer_ids = settings_dict.get(enums.OptimizerConfig.OPTIMIZER_IDS.value)
        self.optimizer_mode = settings_dict.get(enums.OptimizerConfig.MODE.value,
                                                enums.OptimizerModes.NORMAL.value)
        self.queue_size = int(settings_dict.get(enums.OptimizerConfig.QUEUE_SIZE.value,
                                                constants.OPTIMIZER_DEFAULT_QUEUE_SIZE))
        self.empty_the_queue = settings_dict.get(enums.OptimizerConfig.EMPTY_THE_QUEUE.value, False)
        # update run database at the end of each period
        self.db_update_period = int(settings_dict.get(enums.OptimizerConfig.DB_UPDATE_PERIOD.value,
                                                      constants.OPTIMIZER_DEFAULT_DB_UPDATE_PERIOD))
        # AI / genetic
        self.max_optimizer_runs = settings_dict.get(enums.OptimizerConfig.MAX_OPTIMIZER_RUNS.value,
                                                    constants.OPTIMIZER_DEFAULT_MAX_OPTIMIZER_RUNS)
        self.generations_count = settings_dict.get(enums.OptimizerConfig.DEFAULT_GENERATIONS_COUNT.value,
                                                   constants.OPTIMIZER_DEFAULT_GENERATIONS_COUNT)
        self.initial_generation_count = settings_dict.get(enums.OptimizerConfig.INITIAL_GENERATION_COUNT.value,
                                                          constants.OPTIMIZER_DEFAULT_INITIAL_GENERATION_COUNT)
        self.run_per_generation = settings_dict.get(enums.OptimizerConfig.DEFAULT_RUN_PER_GENERATION.value,
                                                    constants.OPTIMIZER_DEFAULT_RUN_PER_GENERATION)
        self.fitness_parameters = self.parse_fitness_parameters(
            settings_dict.get(enums.OptimizerConfig.DEFAULT_SCORING_PARAMETERS.value,
                              self.get_default_fitness_parameters())
        )
        self.exclude_filters = self.parse_optimizer_filter(
            settings_dict.get(enums.OptimizerConfig.DEFAULT_OPTIMIZER_FILTERS.value,
                              self.get_default_optimizer_filters())
        )
        self.constraints_by_key = self.parse_optimizer_constraint(
            settings_dict.get(enums.OptimizerConfig.DEFAULT_OPTIMIZER_CONSTRAINTS.value,
                              self.get_default_optimizer_constraints())
        )
        self.mutation_percent = float(settings_dict.get(
            enums.OptimizerConfig.DEFAULT_MUTATION_PERCENT.value,
            constants.OPTIMIZER_DEFAULT_MUTATION_PERCENT))
        self.max_mutation_probability_percent = decimal.Decimal(settings_dict.get(
            enums.OptimizerConfig.MAX_MUTATION_PROBABILITY_PERCENT.value,
            constants.OPTIMIZER_DEFAULT_MAX_MUTATION_PROBABILITY_PERCENT))
        self.min_mutation_probability_percent = decimal.Decimal(settings_dict.get(
            enums.OptimizerConfig.MIN_MUTATION_PROBABILITY_PERCENT.value,
            constants.OPTIMIZER_DEFAULT_MIN_MUTATION_PROBABILITY_PERCENT))
        self.max_mutation_number_multiplier = decimal.Decimal(settings_dict.get(
            enums.OptimizerConfig.DEFAULT_MAX_MUTATION_NUMBER_MULTIPLIER.value,
            constants.OPTIMIZER_DEFAULT_MAX_MUTATION_NUMBER_MULTIPLIER))
        self.target_fitness_score = settings_dict.get(enums.OptimizerConfig.TARGET_FITNESS_SCORE.value)
        self.stay_within_boundaries = settings_dict.get(enums.OptimizerConfig.STAY_WITHIN_BOUNDARIES.value,
                                                        False)

    def get_constraint(self, constraint_key):
        if constraint_key in self.constraints_by_key:
            return self.constraints_by_key[constraint_key]
        return None

    def parse_fitness_parameters(self, parameters):
        return [
            fitness_parameter.FitnessParameter.from_dict(param)
            for param in parameters
        ]

    def get_default_fitness_parameters(self):
        return [
            {
                fitness_parameter.FitnessParameter.NAME_KEY: commons_enums.BacktestingMetadata.PERCENT_GAINS.value,
                fitness_parameter.FitnessParameter.WEIGHT_KEY: 1,
                fitness_parameter.FitnessParameter.IS_RATIO_FROM_MAX_KEY: True,
            },
            {
                fitness_parameter.FitnessParameter.NAME_KEY:
                    commons_enums.BacktestingMetadata.COEFFICIENT_OF_DETERMINATION_MAX_BALANCE.value,
                fitness_parameter.FitnessParameter.WEIGHT_KEY: 0,
                fitness_parameter.FitnessParameter.IS_RATIO_FROM_MAX_KEY: False,
            },
        ]

    def parse_optimizer_filter(self, filters):
        return [
            optimizer_filter.OptimizerFilter.from_dict(element)
            for element in filters
        ]

    def get_default_optimizer_filters(self):
        return [
            {
                optimizer_filter.OptimizerFilter.LEFT_OPERAND_KEY_KEY: commons_enums.BacktestingMetadata.TRADES.value,
                optimizer_filter.OptimizerFilter.RIGHT_OPERAND_KEY_KEY: None,
                optimizer_filter.OptimizerFilter.LEFT_OPERAND_VALUE_KEY: None,
                optimizer_filter.OptimizerFilter.RIGHT_OPERAND_VALUE_KEY: 1,
                optimizer_filter.OptimizerFilter.OPERATOR_KEY: commons_enums.LogicalOperators.LOWER_THAN.value,
            },
            {
                optimizer_filter.OptimizerFilter.LEFT_OPERAND_KEY_KEY:
                    commons_enums.BacktestingMetadata.COEFFICIENT_OF_DETERMINATION_MAX_BALANCE.value,
                optimizer_filter.OptimizerFilter.RIGHT_OPERAND_KEY_KEY: None,
                optimizer_filter.OptimizerFilter.LEFT_OPERAND_VALUE_KEY: None,
                optimizer_filter.OptimizerFilter.RIGHT_OPERAND_VALUE_KEY: 0,
                optimizer_filter.OptimizerFilter.OPERATOR_KEY: commons_enums.LogicalOperators.LOWER_THAN.value,
            },
            {
                optimizer_filter.OptimizerFilter.LEFT_OPERAND_KEY_KEY:
                    commons_enums.BacktestingMetadata.PERCENT_GAINS.value,
                optimizer_filter.OptimizerFilter.RIGHT_OPERAND_KEY_KEY: None,
                optimizer_filter.OptimizerFilter.LEFT_OPERAND_VALUE_KEY: None,
                optimizer_filter.OptimizerFilter.RIGHT_OPERAND_VALUE_KEY: 0,
                optimizer_filter.OptimizerFilter.OPERATOR_KEY: commons_enums.LogicalOperators.LOWER_THAN.value,
            },
        ]

    def parse_optimizer_constraint(self, constaints):
        return {
            element.name: element
            for element in (
                optimizer_constraint.OptimizerConstraint.from_dict(element)
                for element in constaints
            )
        }

    def get_default_optimizer_constraints(self):
        # return [
        #     {
        #         optimizer_constraint.OptimizerConstraint.NAME_KEY: "tentacle-user_input",
        #         optimizer_constraint.OptimizerConstraint.MIN_VAL_KEY: 1,
        #         optimizer_constraint.OptimizerConstraint.MAX_VAL_KEY: 3,
        #         optimizer_constraint.OptimizerConstraint.MIN_STEP_KEY: 0.25,
        #         optimizer_constraint.OptimizerConstraint.MAX_STEP_KEY: 1,
        #         optimizer_constraint.OptimizerConstraint.STAY_WITHIN_BOUNDARIES_KEY: False,
        #     }
        # ]
        return []



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


class OptimizerConstraint:
    NAME_KEY = "name"
    MIN_VAL_KEY = "min_val"
    MAX_VAL_KEY = "max_val"
    MIN_STEP_KEY = "min_step"
    MAX_STEP_KEY = "max_step"
    STAY_WITHIN_BOUNDARIES_KEY = "stay_within_boundaries"

    def __init__(self, name, min_val, max_val, min_step, max_step, stay_within_boundaries):
        self.name = name
        self.min_val = min_val
        self.max_val = max_val
        self.min_step = min_step
        self.max_step = max_step
        self.stay_within_boundaries = stay_within_boundaries

    def is_min_max_valid(self, value):
        if self.max_val is not None and not value > self.max_val:
            return False
        if self.min_val is not None and not value < self.min_val:
            return False
        return True

    @classmethod
    def from_dict(cls, param_dict):
        return cls(
            param_dict.get(cls.NAME_KEY),
            param_dict.get(cls.MIN_VAL_KEY),
            param_dict.get(cls.MAX_VAL_KEY),
            param_dict.get(cls.MIN_STEP_KEY),
            param_dict.get(cls.MAX_STEP_KEY),
            param_dict.get(cls.STAY_WITHIN_BOUNDARIES_KEY),
        )



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import mock

import octobot.automation
import octobot_commons.configuration as configuration


class TestTriggerEvent(octobot.automation.AbstractTriggerEvent):
    def __init__(self):
        super().__init__()
        self.get_next_event_mock = mock.AsyncMock()
        self.trigger_only_once = True

    async def _get_next_event(self):
        # trigger instantly
        await self.get_next_event_mock()

    @staticmethod
    def get_description() -> str:
        return ""

    def get_user_inputs(self, UI: configuration.UserInputFactory, inputs: dict, step_name: str) -> dict:
        return {}

    def apply_config(self, config):
        return


class TestCondition(octobot.automation.AbstractCondition):
    def __init__(self):
        super().__init__()
        self.evaluate_mock = mock.AsyncMock()

    async def evaluate(self) -> bool:
        await self.evaluate_mock()
        return True

    @staticmethod
    def get_description() -> str:
        return ""

    def get_user_inputs(self, UI: configuration.UserInputFactory, inputs: dict, step_name: str) -> dict:
        return {}

    def apply_config(self, config):
        return


class TestAction(octobot.automation.AbstractAction):
    def __init__(self):
        super().__init__()
        self.process_mock = mock.AsyncMock()

    async def process(self):
        # trigger instantly
        await self.process_mock()

    @staticmethod
    def get_description() -> str:
        return ""

    def get_user_inputs(self, UI: configuration.UserInputFactory, inputs: dict, step_name: str) -> dict:
        return {}

    def apply_config(self, config):
        return



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import octobot_backtesting.api as backtesting_api

import octobot_commons.enums as common_enums

import octobot_services.api as service_api
import octobot_services.octobot_channel_consumer as service_channel_consumer

import octobot_tentacles_manager.api as tentacles_manager_api

import octobot.channels as octobot_channels
import octobot.constants as constants


class ServiceFeedProducer(octobot_channels.OctoBotChannelProducer):
    """EvaluatorFactory class:
    - Create service feeds
    """

    def __init__(self, channel, octobot):
        super().__init__(channel)
        self.octobot = octobot
        self.started = False

        self.service_feeds = []

    async def start(self):
        in_backtesting = backtesting_api.is_backtesting_enabled(self.octobot.config)
        service_feed_factory = service_api.create_service_feed_factory(self.octobot.config,
                                                                       self.octobot.async_loop,
                                                                       self.octobot.bot_id)
        for feed in service_feed_factory.get_available_service_feeds(in_backtesting):
            if tentacles_manager_api.is_tentacle_activated_in_tentacles_setup_config(
                    self.octobot.tentacles_setup_config, feed.get_name()):
                await self.create_feed(service_feed_factory, feed, in_backtesting)

    async def start_feeds(self):
        self.started = True
        for feed in self.service_feeds:
            await self.send(bot_id=self.octobot.bot_id,
                            subject=common_enums.OctoBotChannelSubjects.UPDATE.value,
                            action=service_channel_consumer.OctoBotChannelServiceActions.START_SERVICE_FEED.value,
                            data={
                                service_channel_consumer.OctoBotChannelServiceDataKeys.INSTANCE.value: feed,
                                service_channel_consumer.OctoBotChannelServiceDataKeys.EDITED_CONFIG.value:
                                    self.octobot.get_edited_config(constants.CONFIG_KEY, dict_only=False)
                            })

    async def create_feed(self, service_feed_factory, feed, in_backtesting):
        await self.send(bot_id=self.octobot.bot_id,
                        subject=common_enums.OctoBotChannelSubjects.CREATION.value,
                        action=service_channel_consumer.OctoBotChannelServiceActions.SERVICE_FEED.value,
                        data={
                            service_channel_consumer.OctoBotChannelServiceDataKeys.EDITED_CONFIG.value:
                                self.octobot.get_edited_config(constants.CONFIG_KEY, dict_only=False),
                            service_channel_consumer.OctoBotChannelServiceDataKeys.BACKTESTING_ENABLED.value:
                                in_backtesting,
                            service_channel_consumer.OctoBotChannelServiceDataKeys.CLASS.value: feed,
                            service_channel_consumer.OctoBotChannelServiceDataKeys.FACTORY.value: service_feed_factory
                        })

    async def register_service_feed(self, instance):
        self.service_feeds.append(instance)

    async def stop(self):
        self.logger.debug("Stopping ...")
        for service_feed in self.service_feeds:
            await service_api.stop_service_feed(service_feed)
        self.logger.debug("Stopped")



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

import octobot_commons.os_util as os_util
import octobot_commons.enums as commons_enums

import octobot.updater.binary_updater as binary_updater
import octobot.updater.python_updater as python_updater


def create_updater():
    bot_type = os_util.get_octobot_type()

    if bot_type == commons_enums.OctoBotTypes.DOCKER.value:
        return None
    if bot_type == commons_enums.OctoBotTypes.BINARY.value:
        return binary_updater.BinaryUpdater()
    if bot_type == commons_enums.OctoBotTypes.PYTHON.value:
        return python_updater.PythonUpdater()
    return None



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import octobot_commons.logical_operators as logical_operators


class OptimizerFilter:
    LEFT_OPERAND_KEY_KEY = "left_operand_key"
    RIGHT_OPERAND_KEY_KEY = "right_operand_key"
    LEFT_OPERAND_VALUE_KEY = "left_operand_value"
    RIGHT_OPERAND_VALUE_KEY = "right_operand_value"
    OPERATOR_KEY = "operator"

    def __init__(self, left_operand_key, right_operand_key, left_operand_value, right_operand_value, operator):
        self.left_operand_key = left_operand_key
        self.right_operand_key = right_operand_key
        self.left_operand_value = left_operand_value
        self.right_operand_value = right_operand_value
        self.operator = operator

    def is_valid(self):
        return self.left_operand_value and self.right_operand_value and self.operator

    def load_values(self, values: dict):
        succeeded = False
        if self.left_operand_key is not None:
            try:
                self.left_operand_value = values[self.left_operand_key]
                succeeded = True
            except KeyError:
                pass
        if self.right_operand_key is not None:
            try:
                self.right_operand_value = values[self.right_operand_key]
            except KeyError:
                if not succeeded:
                    # require at least one value read
                    raise

    def is_filtered(self):
        if not self.is_valid():
            return False
        try:
            left_operand = decimal.Decimal(self.left_operand_value)
        except decimal.InvalidOperation:
            left_operand = str(self.left_operand_value)
        try:
            right_operand = decimal.Decimal(self.right_operand_value)
        except decimal.InvalidOperation:
            right_operand = str(self.right_operand_value)
        return logical_operators.evaluate_condition(left_operand, right_operand, self.operator)

    @classmethod
    def from_dict(cls, param_dict):
        return cls(
            param_dict[cls.LEFT_OPERAND_KEY_KEY],
            param_dict[cls.RIGHT_OPERAND_KEY_KEY],
            param_dict[cls.LEFT_OPERAND_VALUE_KEY],
            param_dict[cls.RIGHT_OPERAND_VALUE_KEY],
            param_dict[cls.OPERATOR_KEY],
        )



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import typing
import gotrue
import octobot_commons.configuration
import octobot_commons.logging
import octobot.constants


class SyncConfigurationStorage(gotrue.SyncSupportedStorage):
    """
    Used by gotrue client to save authenticated user session
    """
    def __init__(self, configuration: octobot_commons.configuration.Configuration):
        self.configuration: octobot_commons.configuration.Configuration = configuration

    def get_item(self, key: str) -> typing.Optional[str]:
        return self._get_value_in_config(key)

    def set_item(self, key: str, value: str) -> None:
        self._save_value_in_config(key, value)

    def remove_item(self, key: str) -> None:
        self._save_value_in_config(key, "")

    def _save_value_in_config(self, key, value):
        if self.configuration is not None:
            if octobot.constants.CONFIG_COMMUNITY not in self.configuration.config:
                self.configuration.config[octobot.constants.CONFIG_COMMUNITY] = {}
            self.configuration.config[octobot.constants.CONFIG_COMMUNITY][key] = value
            try:
                self.configuration.save()
            except Exception as err:
                octobot_commons.logging.get_logger(self.__class__.__name__).exception(
                    err, True, f"Error when saving configuration {err}"
                )

    def _get_value_in_config(self, key):
        if self.configuration is not None:
            try:
                return self.configuration.config[octobot.constants.CONFIG_COMMUNITY][key]
            except KeyError:
                return ""
        return None


class ASyncConfigurationStorage(gotrue.AsyncSupportedStorage):
    def __init__(self, configuration: octobot_commons.configuration.Configuration):
        self.sync_storage: SyncConfigurationStorage = SyncConfigurationStorage(configuration)

    async def get_item(self, key: str) -> typing.Optional[str]:
        return self.sync_storage.get_item(key)

    async def set_item(self, key: str, value: str) -> None:
        self.sync_storage.set_item(key, value)

    async def remove_item(self, key: str) -> None:
        self.sync_storage.remove_item(key)



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio
import time
import uuid
import aiohttp

import octobot_commons.constants as commons_constants
import octobot_commons.enums as commons_enums
import octobot_commons.logging as logging
import octobot_commons.configuration as configuration
import octobot_commons.signals as signals
import octobot_commons.databases as databases
import octobot_commons.tree as commons_tree
import octobot_commons.os_clock_sync as os_clock_sync
import octobot_commons.system_resources_watcher as system_resources_watcher

import octobot_services.api as service_api
import octobot_trading.api as trading_api

import octobot.logger as logger
import octobot.community as community
import octobot.constants as constants
import octobot.configuration_manager as configuration_manager
import octobot.task_manager as task_manager
import octobot.octobot_channel_consumer as octobot_channel_consumer
import octobot.octobot_api as octobot_api
import octobot.initializer as initializer
import octobot.producers as producers
import octobot.storage as storage
import octobot.automation as automation

"""Main OctoBot class:
- Create all indicators and thread for each cryptocurrencies in config """


class OctoBot:
    """
    Constructor :
    - Load configs
    """

    def __init__(self, config: configuration.Configuration, community_authenticator=None,
                 ignore_config=False, reset_trading_history=False, startup_messages=None):
        self.start_time = time.time()
        self.config = config.config
        self.ignore_config = ignore_config
        self.reset_trading_history = reset_trading_history
        self.startup_messages = startup_messages

        # tentacle setup configuration
        self.tentacles_setup_config = None

        # Configuration manager to handle current, edited and startup configurations
        self.configuration_manager = configuration_manager.ConfigurationManager()
        self.configuration_manager.add_element(constants.CONFIG_KEY, config, has_dict=True)

        # Used to know when OctoBot is ready to answer in APIs
        self.initialized = False

        # unique aiohttp session: to be initialized from getter in a task
        self._aiohttp_session = None

        # community if enabled
        self.community_handler = None

        # use edited config in community authentication
        community_config = self.get_edited_config(constants.CONFIG_KEY, dict_only=False)
        self.community_auth = community_authenticator or community.CommunityAuthentication.create(community_config)
        self.community_auth.update(community_config)

        # octobot_api to request the current instance
        self.octobot_api = octobot_api.OctoBotAPI(self)

        # octobot channel global consumer
        self.global_consumer = octobot_channel_consumer.OctoBotChannelGlobalConsumer(self)

        # octobot instance id
        self.bot_id = str(uuid.uuid4())

        # Logger
        self.logger = logging.get_logger(self.__class__.__name__)

        # automations
        self.automation = None

        # Initialize octobot main tools
        self.initializer = initializer.Initializer(self)
        self.task_manager = task_manager.TaskManager(self)
        self._init_metadata_run_task = None

        # Producers
        self.exchange_producer = None
        self.evaluator_producer = None
        self.interface_producer = None
        self.service_feed_producer = None

        self.async_loop = None
        self.stopped = None

    async def initialize(self):
        self.stopped = asyncio.Event()
        await self._ensure_clock()
        if not (self.community_auth.is_initialized() and self.community_auth.is_using_the_current_loop()):
            self.community_auth.init_account()
        self._log_config()
        await self.initializer.create(True)
        await self._start_tools_tasks()
        await logger.init_octobot_chan_logger(self.bot_id)
        await self.create_producers()
        await self.start_producers()
        await self._ensure_watchers()
        await self._post_initialize()

    async def create_producers(self):
        self.exchange_producer = producers.ExchangeProducer(self.global_consumer.octobot_channel, self,
                                                            None, self.ignore_config)
        self.evaluator_producer = producers.EvaluatorProducer(self.global_consumer.octobot_channel, self)
        self.interface_producer = producers.InterfaceProducer(self.global_consumer.octobot_channel, self)
        self.service_feed_producer = producers.ServiceFeedProducer(self.global_consumer.octobot_channel, self)

    async def start_producers(self):
        await self.evaluator_producer.run()
        await self.exchange_producer.run()
        # Start service feeds now that evaluators registered their feed requirements
        await self.service_feed_producer.run()
        await self.interface_producer.run()

    async def _post_initialize(self):
        self.initialized = True

        # make tentacles setup config editable while saving previous states
        self.configuration_manager.add_element(constants.TENTACLES_SETUP_CONFIG_KEY, self.tentacles_setup_config)
        await service_api.send_notification(
            service_api.create_notification(f"{constants.PROJECT_NAME} {constants.LONG_VERSION} is starting ...",
                                            markdown_format=commons_enums.MarkdownFormat.ITALIC)
        )
        if self.startup_messages:
            for limit_message in self.startup_messages:
                self.logger.info(f"Startup message: {limit_message}")
                await service_api.send_notification(
                    service_api.create_notification(limit_message)
                )

        self.automation = automation.Automation(self.bot_id, self.tentacles_setup_config)
        self._init_metadata_run_task = asyncio.create_task(self._store_run_metadata_when_available())

    async def _wait_for_run_data_init(self, exchange_managers, timeout):
        for exchange_manager in exchange_managers:
            for topic in constants.REQUIRED_TOPIC_FOR_DATA_INIT:
                await commons_tree.EventProvider.instance().wait_for_event(
                    self.bot_id,
                    commons_tree.get_exchange_path(
                        trading_api.get_exchange_name(exchange_manager),
                        topic.value
                    ),
                    timeout
                )

    async def _store_run_metadata_when_available(self):
        run_metadata_init_timeout = 5 * commons_constants.MINUTE_TO_SECONDS
        # first wait for all exchanges to be created
        try:
            await asyncio.wait_for(self.exchange_producer.created_all_exchanges.wait(), run_metadata_init_timeout)
        except asyncio.TimeoutError:
            pass
        exchange_managers = [
            trading_api.get_exchange_manager_from_exchange_id(exchange_manager_id)
            for exchange_manager_id in self.exchange_producer.exchange_manager_ids
            if trading_api.is_trader_existing_and_enabled(
                trading_api.get_exchange_manager_from_exchange_id(exchange_manager_id)
            )
        ]
        # start automations now that everything started
        await self.automation.initialize()
        try:
            await self._wait_for_run_data_init(exchange_managers, run_metadata_init_timeout)
        except asyncio.TimeoutError:
            pass
        try:
            if exchange_managers:
                await storage.clear_run_metadata(self.bot_id)
                await storage.store_run_metadata(self.bot_id, exchange_managers, self.start_time, flush=True)
            else:
                self.logger.debug("Skipping run metadata update: no available exchange manager")
        except Exception as err:
            self.logger.exception(err, True, f"Error when storing live metadata: {err}")

    async def stop(self):
        try:
            self.logger.debug("Stopping ...")
            if self._init_metadata_run_task is not None and not self._init_metadata_run_task.done():
                self._init_metadata_run_task.cancel()
            signals.SignalPublisher.instance().stop()
            await self.evaluator_producer.stop()
            await self.exchange_producer.stop()
            await self.community_auth.stop()
            await self.service_feed_producer.stop()
            await os_clock_sync.stop_clock_synchronizer()
            await system_resources_watcher.stop_system_resources_watcher()
            await service_api.stop_services()
            await self.interface_producer.stop()
            await databases.close_bot_storage(self.bot_id)
            if self.automation is not None:
                await self.automation.stop()

        finally:
            self.stopped.set()
            self.logger.info("Stopped, now shutting down.")

    async def _start_tools_tasks(self):
        self._init_community()
        await self.task_manager.start_tools_tasks()

    def _init_community(self):
        self.community_handler = community.CommunityManager(self.octobot_api)

    async def _ensure_clock(self):
        if trading_api.is_trader_enabled_in_config(self.config) and constants.ENABLE_CLOCK_SYNCH:
            await os_clock_sync.start_clock_synchronizer()

    async def _ensure_watchers(self):
        if constants.ENABLE_SYSTEM_WATCHER:
            await system_resources_watcher.start_system_resources_watcher(
                constants.DUMP_USED_RESOURCES,
                constants.USED_RESOURCES_OUTPUT
            )

    def _log_config(self):
        exchanges = [
            f"{exchange}" \
            f"[{config.get(commons_constants.CONFIG_EXCHANGE_TYPE, trading_api.get_default_exchange_type(exchange))}]"
            for exchange, config in self.config.get(commons_constants.CONFIG_EXCHANGES, {}).items()
            if config.get(commons_constants.CONFIG_ENABLED_OPTION, True)
        ]
        has_real_trader = trading_api.is_trader_enabled_in_config(self.config)
        has_simulated_trader = trading_api.is_trader_simulator_enabled_in_config(self.config)
        trader_str = "real trader" if has_real_trader else "simulated trader" if has_simulated_trader else "no trader"
        traded_symbols = trading_api.get_config_symbols(self.config, True)
        symbols_str = ', '.join(set(traded_symbols))
        self.logger.info(f"Starting OctoBot with {trader_str} on "
                         f"{', '.join(exchanges) if exchanges else 'no exchange'} "
                         f"trading {symbols_str or 'nothing'} and using bot_id: {self.bot_id}")

    def get_edited_config(self, config_key, dict_only=True):
        return self.configuration_manager.get_edited_config(config_key, dict_only)

    def set_edited_config(self, config_key, config):
        self.configuration_manager.set_edited_config(config_key, config)

    def get_startup_config(self, config_key, dict_only=True):
        return self.configuration_manager.get_startup_config(config_key, dict_only)

    def get_trading_mode(self):
        try:
            first_exchange_manager = trading_api.get_exchange_manager_from_exchange_id(
                next(iter(self.exchange_producer.exchange_manager_ids))
            )
            return trading_api.get_trading_modes(first_exchange_manager)[0]
        except (StopIteration, IndexError):
            return None

    def run_in_main_asyncio_loop(self, coroutine, log_exceptions=True,
                                 timeout=commons_constants.DEFAULT_FUTURE_TIMEOUT):
        return self.task_manager.run_in_main_asyncio_loop(coroutine, log_exceptions=log_exceptions, timeout=timeout)

    def set_watcher(self, watcher):
        self.task_manager.watcher = watcher

    def get_aiohttp_session(self):
        if self._aiohttp_session is None:
            self._aiohttp_session = aiohttp.ClientSession()
        return self._aiohttp_session



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.

import enum


class BotKeys(enum.Enum):
    ID = "id"
    CREATED_AT = "created_at"
    NAME = "name"
    USER_ID = "user_id"
    CURRENT_PORTFOLIO_ID = "current_portfolio_id"
    CURRENT_DEPLOYMENT_ID = "current_deployment_id"
    CURRENT_CONFIG_ID = "current_config_id"
    LAST_TRADE_TIME = "last_trade_time"
    METADATA = "metadata"
    ORDERS = "orders"


class ProductsSubscriptionsKeys(enum.Enum):
    ID = "id"
    CREATED_AT = "created_at"
    STATUS = "status"
    DESIRED_STATUS = "desired_status"


class BotDeploymentKeys(enum.Enum):
    ID = "id"
    CREATED_AT = "created_at"
    NAME = "name"
    VERSION = "version"
    BOT_ID = "bot_id"
    CURRENT_HOSTING_ID = "current_hosting_id"
    PRODUCT_ID = "product_id"
    CURRENT_STORAGE_ID = "current_storage_id"
    SUBSCRIPTION_ID = "product_subscription_id"
    CURRENT_URL_ID = "current_url_id"
    STATUS = "status"
    DESIRED_STATUS = "desired_status"
    TYPE = "type"
    METADATA = "metadata"
    ERROR_STATUS = "error_status"
    ACTIVITIES = "activities"


class BotDeploymentActivitiesKeys(enum.Enum):
    LAST_ACTIVITY = "last_activity"
    NEXT_ACTIVITY = "next_activity"


class BotDeploymentURLKeys(enum.Enum):
    ID = "id"
    CREATED_AT = "created_at"
    REGION = "region"
    DEPLOYMENT_ID = "deployment_id"
    URL = "url"


class BotDeploymentStatus(enum.Enum):
    CREATED = 'created'
    UPDATED = "updated"
    RUNNING = "running"
    STOPPED = "stopped"
    PENDING = "pending"
    UNKNOWN = "unknown"


class ProductSubscriptionDesiredStatus(enum.Enum):
    ACTIVE = 'active'
    CANCELED = "canceled"


class BotDeploymentErrorsStatuses(enum.Enum):
    NO_ERROR = None
    INTERNAL_SERVER_ERROR = "internal_server_error"
    INVALID_CONFIG = "invalid_config"
    INVALID_EXCHANGE_CREDENTIALS = "invalid_exchange_credentials"
    ALREADY_USED_EXCHANGE_ACCOUNT = "already_used_exchange_account"
    MISSING_MINIMAL_FUNDS = "missing_minimal_funds"
    MISSING_CONFIG = "missing_config"


class ExchangeAccountStatuses(enum.Enum):
    NO_STATUS = None
    PENDING_PORTFOLIO_REFRESH = "pending_portfolio_refresh"
    NO_PENDING_ACTION = "no_pending_action"


class SignalKeys(enum.Enum):
    ID = "id"
    TIME = "time"
    PRODUCT_ID = "product_id"
    SIGNAL = "signal"


class ProductKeys(enum.Enum):
    ID = "id"
    CREATED_AT = "created_at"
    CONTENT = "content"
    UPDATED_AT = "updated_at"
    SLUG = "slug"
    CURRENT_VERSION_ID = "current_version_id"
    CURRENT_CONFIG_ID = "current_config_id"
    AUTHOR_ID = "author_id"
    LOGO_URL = "logo_url"
    DOWNLOAD_URL = "download_url"
    CATEGORY_ID = "category_id"
    ATTRIBUTES = "attributes"
    METADATA = "metadata"
    PARENT_ID = "parent_id"
    VISIBILITY = "visibility"
    CURRENT_RESULT_ID = "current_result_id"


class TradeKeys(enum.Enum):
    ID = "id"
    TRADE_ID = "trade_id"
    BOT_ID = "bot_id"
    TIME = "time"
    EXCHANGE = "exchange"
    PRICE = "price"
    QUANTITY = "quantity"
    SYMBOL = "symbol"
    TYPE = "type"
    VOLUME = "volume"
    METADATA = "metadata"


class OrderKeys(enum.Enum):
    TIME = "time"
    EXCHANGE = "exchange"
    PRICE = "price"
    QUANTITY = "quantity"
    SYMBOL = "symbol"
    TYPE = "type"


class PortfolioKeys(enum.Enum):
    ID = "id"
    BOT_ID = "bot_id"
    CURRENT_VALUE = "current_value"
    INITIAL_VALUE = "initial_value"
    PROFITABILITY = "profitability"
    UNIT = "unit"
    CONTENT = "content"


class PortfolioHistoryKeys(enum.Enum):
    TIME = "time"
    PORTFOLIO_ID = "portfolio_id"
    VALUE = "value"


class PortfolioAssetKeys(enum.Enum):
    ASSET = "asset"
    VALUE = "value"
    QUANTITY = "quantity"


class BotConfigKeys(enum.Enum):
    ID = "id"
    BOT_ID = "bot_id"
    PRODUCT_CONFIG_ID = "product_config_id"
    EXCHANGES = "exchanges"
    OPTIONS = "options"


class ProfileConfigKeys(enum.Enum):
    ID = "id"
    PRODUCT_CONFIG_ID = "product_id"
    VERSION = "version"
    CONFIG = "config"


class UserKeys(enum.Enum):
    ID = "id"
    EMAIL = "email"
    USER_METADATA = "user_metadata"
    LAST_SIGN_IN_AT = "last_sign_in_at"


class DeploymentTypes(enum.Enum):
    SELF_HOSTED = "self_hosted"
    CLOUD = "cloud"


class SQLValues(enum.Enum):
    NULL = "null"



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import pytest
import mock

import octobot.constants as constants
import octobot.community as community
import octobot.community.errors_upload.initializer as initializer
import octobot_commons.logging as logging
import octobot_commons.logging.logging_util as logging_util
from tests.unit_tests.community.errors_upload import UPLOADER_URL


@pytest.fixture()
def upload_wrapper():
    config = mock.Mock()
    config.get_metrics_id = mock.Mock(return_value="abc")
    return initializer._UploadWrapper(UPLOADER_URL, config)


def test_register_error_uploader():
    config = mock.Mock()
    upload_wrapper = mock.Mock()
    upload_wrapper.upload_if_necessary = "upload_if_necessary"
    previous_callback = logging_util._ERROR_CALLBACK
    with mock.patch.object(initializer, "_UploadWrapper", mock.Mock(return_value=upload_wrapper)) \
            as _UploadWrapper_mock:
        community.register_error_uploader(UPLOADER_URL, config)
        _UploadWrapper_mock.assert_called_once_with(UPLOADER_URL, config)
        assert logging_util._ERROR_CALLBACK == "upload_if_necessary"
    # restore callback
    logging.BotLogger.register_error_callback(previous_callback)


def test_UploadWrapper_upload_if_necessary(upload_wrapper):
    with mock.patch.object(constants, "UPLOAD_ERRORS", True):
        with mock.patch.object(upload_wrapper._uploader, "schedule_error_upload", mock.Mock()) \
                as schedule_error_upload_mock:
            with mock.patch.object(upload_wrapper._config, "get_metrics_enabled", mock.Mock(return_value=False)) \
                 as get_metrics_enabled_mock:
                upload_wrapper.upload_if_necessary(None, "message")
                get_metrics_enabled_mock.assert_called_once()
                schedule_error_upload_mock.assert_not_called()
            with mock.patch.object(upload_wrapper._config, "get_metrics_enabled", mock.Mock(return_value=True)) \
                 as get_metrics_enabled_mock:
                upload_wrapper.upload_if_necessary(None, "message")
                get_metrics_enabled_mock.assert_called_once()
                schedule_error_upload_mock.assert_called_once()
    with mock.patch.object(constants, "UPLOAD_ERRORS", False):
        with mock.patch.object(upload_wrapper._uploader, "schedule_error_upload", mock.Mock()) \
                as schedule_error_upload_mock:
            with mock.patch.object(upload_wrapper._config, "get_metrics_enabled", mock.Mock(return_value=True)) \
                 as get_metrics_enabled_mock:
                upload_wrapper.upload_if_necessary(None, "message")
                get_metrics_enabled_mock.assert_not_called()
                schedule_error_upload_mock.assert_not_called()


def test_get_metrics_id(upload_wrapper):
    def raiser():
        return {}["plop"]
    assert upload_wrapper._get_metrics_id() == "abc"
    with mock.patch.object(upload_wrapper._config, "get_metrics_id", raiser):
        assert upload_wrapper._get_metrics_id() == constants.DEFAULT_METRICS_ID



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio

from octobot_backtesting.errors import MissingTimeFrame
from octobot_commons.logging.logging_util import get_logger
from octobot_commons.tests.test_config import load_test_config

from octobot.api.backtesting import create_independent_backtesting, \
    initialize_and_run_independent_backtesting, stop_independent_backtesting
from octobot.logger import init_logger
from octobot.octobot import OctoBot
from tests.test_utils.config import load_test_tentacles_config


async def create_bot() -> OctoBot:
    return OctoBot(load_test_config())


async def initialize_bot(bot):
    await bot.initialize()


async def run_independent_backtesting(data_files, timeout=10, use_loggers=True, run_on_common_part_only=True,
                                      enable_storage=False):
    independent_backtesting = None
    try:
        config_to_use = load_test_config()
        if use_loggers:
            init_logger()
        independent_backtesting = create_independent_backtesting(config_to_use,
                                                                 load_test_tentacles_config(),
                                                                 data_files,
                                                                 "",
                                                                 run_on_common_part_only=run_on_common_part_only,
                                                                 enable_storage=enable_storage)
        await initialize_and_run_independent_backtesting(independent_backtesting, log_errors=False)
        await independent_backtesting.join_backtesting_updater(timeout)
        return independent_backtesting
    except MissingTimeFrame:
        # ignore this exception: is due to missing of the only required time frame
        return independent_backtesting
    except asyncio.TimeoutError as e:
        get_logger().exception(e, True, f"Timeout after waiting for backtesting for {timeout} seconds.")
        # stop backtesting to prevent zombie tasks
        await stop_independent_backtesting(independent_backtesting)
        raise
    except Exception as e:
        get_logger().exception(e, True, str(e))
        # stop backtesting to prevent zombie tasks
        await stop_independent_backtesting(independent_backtesting)
        raise



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import asyncio
import copy
import logging
import os.path as path

import octobot_commons.constants as common_constants
import octobot_commons.enums as enums
import octobot_commons.errors as errors
import octobot_commons.logging as commons_logging
import octobot_commons.pretty_printer as pretty_printer
import octobot_commons.symbols.symbol_util as symbol_util
import octobot_commons.databases as databases
import octobot_commons.optimization_campaign as optimization_campaign
import octobot_commons.time_frame_manager as time_frame_manager

import octobot.backtesting as backtesting
import octobot_backtesting.api as backtesting_api
import octobot_backtesting.constants as backtesting_constants
import octobot_backtesting.enums as backtesting_enums
import octobot_backtesting.data as backtesting_data

import octobot_evaluators.constants as evaluator_constants

import octobot_trading.api as trading_api
import octobot_trading.enums as trading_enums

import octobot.storage as storage


class IndependentBacktesting:
    def __init__(
        self,
        config,
        tentacles_setup_config,
        backtesting_files,
        data_file_path=backtesting_constants.BACKTESTING_FILE_PATH,
        run_on_common_part_only=True,
        join_backtesting_timeout=backtesting_constants.BACKTESTING_DEFAULT_JOIN_TIMEOUT,
        start_timestamp=None,
        end_timestamp=None,
        enable_logs=True,
        stop_when_finished=False,
        name=None,
        enforce_total_databases_max_size_after_run=True,
        enable_storage=True,
        run_on_all_available_time_frames=False,
        backtesting_data=None,
        config_by_tentacle=None,
        services_config=None,
    ):
        self.octobot_origin_config = config
        self.tentacles_setup_config = tentacles_setup_config
        self.backtesting_config = {}
        self.backtesting_files = backtesting_files
        self.logger = commons_logging.get_logger(self.__class__.__name__)
        self.data_file_path = data_file_path
        self.symbols_to_create_exchange_classes = {}
        self.risk = 0.1
        self.starting_portfolio = {}
        self.fees_config = {}
        self.forced_time_frames = []
        self.optimizer_id = None
        self.backtesting_id = None
        self._init_default_config_values()
        self.stopped = False
        self.stopped_event = None
        self.post_backtesting_task = None
        self.join_backtesting_timeout = join_backtesting_timeout
        self.enable_logs = enable_logs
        self.stop_when_finished = stop_when_finished
        self.previous_log_level = commons_logging.get_global_logger_level()
        self.previous_handlers_log_level = commons_logging.get_logger_level_per_handler()
        self.enforce_total_databases_max_size_after_run = enforce_total_databases_max_size_after_run
        self.backtesting_data = backtesting_data
        self.octobot_backtesting = backtesting.OctoBotBacktesting(
            self.backtesting_config,
            self.tentacles_setup_config,
            self.symbols_to_create_exchange_classes,
            self.backtesting_files,
            run_on_common_part_only,
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            enable_logs=self.enable_logs,
            enable_storage=enable_storage,
            run_on_all_available_time_frames=run_on_all_available_time_frames,
            backtesting_data=self.backtesting_data,
            name=name,
            config_by_tentacle=config_by_tentacle,
            services_config=services_config,
        )

    async def initialize_and_run(self, log_errors=True):
        try:
            # create stopped_event here only to be sure to create it in the same loop as the one of the
            # backtesting run
            self.logger.debug("Starting backtesting")
            if self.stop_when_finished:
                self.stopped_event = asyncio.Event()
            if not self.enable_logs:
                commons_logging.set_global_logger_level(logging.ERROR)
            commons_logging.reset_backtesting_errors()
            await self.initialize_config()
            await self._generate_backtesting_id_if_missing()
            self._add_crypto_currencies_config()
            await self.octobot_backtesting.initialize_and_run()
            self._post_backtesting_start()
        except RuntimeError as e:
            self.logger.error(f"{e} ({e.__class__.__name__})")
            await self.stop()
        except Exception as e:
            if log_errors:
                self.logger.exception(e, True, f"Error when running backtesting: {e}")
            raise e

    async def initialize_config(self):
        await self._register_available_data()
        self._adapt_config()
        return self.backtesting_config

    async def join_stop_event(self, timeout=None):
        if self.stopped_event is None:
            return
        await asyncio.wait_for(self.stopped_event.wait(), timeout)

    async def join_backtesting_updater(self, timeout=None):
        if self.octobot_backtesting.backtesting is not None:
            await asyncio.wait_for(self.octobot_backtesting.backtesting.time_updater.finished_event.wait(), timeout)

    async def clear_fetched_data(self):
        await self.octobot_backtesting.clear_fetched_data()

    async def stop(self, memory_check=False, should_raise=False):
        try:
            if not self.stopped:
                await self.octobot_backtesting.stop(memory_check=memory_check, should_raise=should_raise)
        finally:
            self.stopped = True
            if self.stopped_event is not None:
                self.stopped_event.set()
            if not self.enable_logs:
                commons_logging.set_global_logger_level(
                    self.previous_log_level,
                    handler_levels=self.previous_handlers_log_level
                )

    def is_in_progress(self):
        if self.octobot_backtesting.backtesting:
            return self.octobot_backtesting.backtesting.is_in_progress()
        else:
            return False

    def has_finished(self):
        if self.octobot_backtesting.backtesting:
            return self.octobot_backtesting.backtesting.has_finished()
        else:
            return False

    def get_progress(self):
        if self.octobot_backtesting.backtesting:
            return self.octobot_backtesting.backtesting.get_progress()
        else:
            return 0

    def _post_backtesting_start(self):
        commons_logging.set_error_publication_enabled(False)
        self.post_backtesting_task = asyncio.create_task(self._register_post_backtesting_end_callback())

    async def _register_post_backtesting_end_callback(self):
        await self.join_backtesting_updater(timeout=self.join_backtesting_timeout)
        await self._post_backtesting_end_callback()

    async def _post_backtesting_end_callback(self):
        # re enable logs
        commons_logging.set_error_publication_enabled(True)
        if self.stop_when_finished:
            await self.stop()
        else:
            # stop backtesting importers to release database files
            await self.octobot_backtesting.stop_importers()
        if self.enforce_total_databases_max_size_after_run:
            try:
                await storage.enforce_total_databases_max_size()
            except Exception as e:
                self.logger.exception(e, True, f"Error when enforcing max run databases size: {e}")

    @staticmethod
    def get_market_delta(symbol, exchange_manager, min_timeframe):
        market_data = trading_api.get_symbol_close_candles(
            trading_api.get_symbol_data(exchange_manager, str(symbol)), min_timeframe
        )
        market_begin = market_data[0]
        market_end = market_data[-1]

        if market_begin and market_end and market_begin > 0:
            market_delta = market_end / market_begin - 1 if market_end >= market_begin \
                else -1 * (1 - market_end / market_begin)
        else:
            market_delta = 0

        return market_delta

    async def _register_available_data(self):
        if not self.symbols_to_create_exchange_classes:
            for data_file in self.backtesting_files:
                data_file_path = data_file
                if not path.isfile(data_file_path):
                    data_file_path = path.join(self.data_file_path, data_file)
                description = await backtesting_data.get_file_description(data_file_path)
                if description is None:
                    raise RuntimeError(f"Impossible to start backtesting: missing or invalid data file: {data_file}")
                exchange_name = description[backtesting_enums.DataFormatKeys.EXCHANGE.value]
                if exchange_name not in self.symbols_to_create_exchange_classes:
                    self.symbols_to_create_exchange_classes[exchange_name] = []
                for symbol in description[backtesting_enums.DataFormatKeys.SYMBOLS.value]:
                    self.symbols_to_create_exchange_classes[exchange_name].append(symbol_util.parse_symbol(symbol))

    def _init_default_config_values(self):
        self.risk = copy.deepcopy(self.octobot_origin_config[common_constants.CONFIG_TRADING][
                                      common_constants.CONFIG_TRADER_RISK])
        self.starting_portfolio = copy.deepcopy(
            self.octobot_origin_config[common_constants.CONFIG_SIMULATOR][common_constants.CONFIG_STARTING_PORTFOLIO]
        )
        self.fees_config = copy.deepcopy(self.octobot_origin_config[common_constants.CONFIG_SIMULATOR][
                                             common_constants.CONFIG_SIMULATOR_FEES])
        if evaluator_constants.CONFIG_FORCED_TIME_FRAME in self.octobot_origin_config:
            self.forced_time_frames = copy.deepcopy(self.octobot_origin_config[
                                                        evaluator_constants.CONFIG_FORCED_TIME_FRAME])
        self.optimizer_id = self.octobot_origin_config.get(common_constants.CONFIG_OPTIMIZER_ID)
        self.backtesting_id = self.octobot_origin_config.get(common_constants.CONFIG_BACKTESTING_ID)
        self.backtesting_config = {
            backtesting_constants.CONFIG_BACKTESTING: {},
            common_constants.CONFIG_CRYPTO_CURRENCIES: {},
            common_constants.CONFIG_EXCHANGES: {},
            common_constants.CONFIG_TRADER: {},
            common_constants.CONFIG_SIMULATOR: {},
            common_constants.CONFIG_TRADING: {},
        }

    async def get_dict_formatted_report(self):
        reference_market = trading_api.get_reference_market(self.backtesting_config)
        try:
            trading_mode = trading_api.get_activated_trading_mode(self.tentacles_setup_config).get_name()
        except errors.ConfigTradingError as e:
            self.logger.error(e)
            trading_mode = "Error when reading trading mode"
        report = self._get_exchanges_report(reference_market, trading_mode)
        return report

    def _get_exchanges_report(self, reference_market, trading_mode):
        SYMBOL_REPORT = "symbol_report"
        BOT_REPORT = "bot_report"
        CHART_IDENTIFIERS = "chart_identifiers"
        ERRORS_COUNT = "errors_count"
        report = {
            SYMBOL_REPORT: [],
            BOT_REPORT: {},
            CHART_IDENTIFIERS: [],
            ERRORS_COUNT: commons_logging.get_backtesting_errors_count()
        }
        profitabilities = {}
        market_average_profitabilities = {}
        starting_portfolios = {}
        end_portfolios = {}
        for exchange_id in self.octobot_backtesting.exchange_manager_ids:
            exchange_manager = trading_api.get_exchange_manager_from_exchange_id(exchange_id)
            _, profitability, _, market_average_profitability, _ = trading_api.get_profitability_stats(exchange_manager)
            min_timeframe = time_frame_manager.find_min_time_frame(trading_api.get_trading_timeframes(exchange_manager))
            exchange_name = trading_api.get_exchange_name(exchange_manager)
            for symbol in self.symbols_to_create_exchange_classes[exchange_name]:
                try:
                    market_delta = self.get_market_delta(symbol, exchange_manager, min_timeframe)
                    report[SYMBOL_REPORT].append({symbol.symbol_str: market_delta * 100})
                    report[CHART_IDENTIFIERS].append({
                        "symbol": symbol.symbol_str,
                        "exchange_id": exchange_id,
                        "exchange_name": exchange_name,
                        "time_frames": list(reversed(
                            [tf.value for tf in trading_api.get_watched_timeframes(exchange_manager)]
                        ))
                    })
                except KeyError:
                    self.logger.debug(f"No price data for {symbol} on {min_timeframe}")
            profitabilities[exchange_name] = float(profitability)
            market_average_profitabilities[exchange_name] = float(market_average_profitability)
            starting_portfolios[exchange_name] = trading_api.get_origin_portfolio(exchange_manager, as_decimal=False)
            end_portfolios[exchange_name] = trading_api.get_portfolio(exchange_manager, as_decimal=False)

        report[BOT_REPORT] = {
            "profitability": profitabilities,
            "market_average_profitability": market_average_profitabilities,
            "reference_market": reference_market,
            "end_portfolio": end_portfolios,
            "starting_portfolio": starting_portfolios,
            "trading_mode": trading_mode
        }
        return report

    def log_report(self):
        self.logger.info(" **** Backtesting report ****")
        for exchange_id in self.octobot_backtesting.exchange_manager_ids:
            exchange_manager = trading_api.get_exchange_manager_from_exchange_id(exchange_id)
            exchange_name = trading_api.get_exchange_name(exchange_manager)
            self.logger.info(
                f" ========= {len(trading_api.get_trade_history(exchange_manager))} Trades on {exchange_name} ========="
            )
            self._log_trades_history(exchange_manager, exchange_name)

            self.logger.info(f" ========= Prices evolution on {exchange_name} =========")
            min_timeframe = time_frame_manager.find_min_time_frame(trading_api.get_trading_timeframes(exchange_manager))
            for symbol in self.symbols_to_create_exchange_classes[exchange_name]:
                self._log_symbol_report(symbol, exchange_manager, min_timeframe)

            self.logger.info(" ========= Octobot end state =========")
            self._log_global_report(exchange_manager)

    def _log_trades_history(self, exchange_manager, exchange_name):
        trades_history_string = "\n".join([pretty_printer.trade_pretty_printer(exchange_name, trade)
                                           for trade in trading_api.get_trade_history(exchange_manager)])
        self.logger.info(f"\n{trades_history_string}")

    def _log_symbol_report(self, symbol, exchange_manager, min_time_frame):
        market_delta = self.get_market_delta(symbol, exchange_manager, min_time_frame)
        self.logger.info(f"{symbol.symbol_str} Profitability : {market_delta * 100}%")

    def _log_global_report(self, exchange_manager):
        _, profitability, _, market_average_profitability, _ = trading_api.get_profitability_stats(exchange_manager)
        reference_market = trading_api.get_reference_market(self.backtesting_config)
        end_portfolio = trading_api.get_portfolio(exchange_manager, as_decimal=False)
        end_portfolio_value = trading_api.get_current_portfolio_value(exchange_manager)
        starting_portfolio = trading_api.get_origin_portfolio(exchange_manager, as_decimal=False)
        starting_portfolio_value = trading_api.get_origin_portfolio_value(exchange_manager)

        self.logger.info(f"[End portfolio]      value {round(end_portfolio_value, 5)} {reference_market} "
                         f"Holdings: {pretty_printer.global_portfolio_pretty_print(end_portfolio, separator=' | ')}")

        self.logger.info(f"[Starting portfolio] value {round(starting_portfolio_value, 5)} {reference_market} "
                         f"Holdings: {pretty_printer.global_portfolio_pretty_print(starting_portfolio, separator=' | ')}")

        self.logger.info(f"Global market profitability (vs {reference_market}) : "
                         f"{market_average_profitability}% | Octobot : {profitability}%")

        self.logger.debug(
            f"Simulation lasted "
            f"{round(backtesting_api.get_backtesting_duration(self.octobot_backtesting.backtesting), 3)} sec")

    def _adapt_config(self):
        self._init_exchange_type()
        self.backtesting_config[common_constants.CONFIG_EXCHANGES] = copy.deepcopy(
            self.octobot_origin_config[common_constants.CONFIG_EXCHANGES])
        for exchange_details in self.backtesting_config[common_constants.CONFIG_EXCHANGES].values():
            exchange_details.pop(common_constants.CONFIG_EXCHANGE_KEY, None)
            exchange_details.pop(common_constants.CONFIG_EXCHANGE_SECRET, None)
            exchange_details.pop(common_constants.CONFIG_EXCHANGE_PASSWORD, None)
            exchange_details.pop(common_constants.CONFIG_EXCHANGE_SANDBOXED, None)
        self.backtesting_config[common_constants.CONFIG_TRADING][common_constants.CONFIG_TRADER_RISK] = self.risk
        self.backtesting_config[common_constants.CONFIG_TRADING][
            common_constants.CONFIG_TRADER_REFERENCE_MARKET] = self._find_reference_market_and_update_contract_type()
        self.backtesting_config[common_constants.CONFIG_SIMULATOR][
            common_constants.CONFIG_STARTING_PORTFOLIO] = self.starting_portfolio
        self.backtesting_config[common_constants.CONFIG_SIMULATOR][
            common_constants.CONFIG_SIMULATOR_FEES] = self.fees_config
        self.backtesting_config[common_constants.CONFIG_OPTIMIZER_ID] = self.optimizer_id
        self.backtesting_config[common_constants.CONFIG_BACKTESTING_ID] = self.backtesting_id
        if self.forced_time_frames:
            self.backtesting_config[evaluator_constants.CONFIG_FORCED_TIME_FRAME] = self.forced_time_frames
        self._add_config_default_backtesting_values()

    def _init_exchange_type(self):
        forced_exchange_type = self.octobot_origin_config.get(common_constants.CONFIG_EXCHANGE_TYPE,
                                                              common_constants.USE_CURRENT_PROFILE)
        try:
            for exchange_name in self.symbols_to_create_exchange_classes:
                if forced_exchange_type == common_constants.USE_CURRENT_PROFILE:
                    # use current profile config to create a spot/future/margin backtesting exchange
                    self.octobot_backtesting.exchange_type_by_exchange[exchange_name] = \
                        self.octobot_origin_config[common_constants.CONFIG_EXCHANGES].get(exchange_name, {}).\
                        get(common_constants.CONFIG_EXCHANGE_TYPE, common_constants.DEFAULT_EXCHANGE_TYPE)
                else:
                    self.octobot_backtesting.exchange_type_by_exchange[exchange_name] = forced_exchange_type
        except StopIteration:
            # use default exchange type
            pass

    async def _generate_backtesting_id_if_missing(self):
        if self.backtesting_config[common_constants.CONFIG_BACKTESTING_ID] is None:
            run_dbs_identifier = databases.RunDatabasesIdentifier(
                trading_api.get_activated_trading_mode(self.tentacles_setup_config),
                optimization_campaign.OptimizationCampaign.get_campaign_name(self.tentacles_setup_config)
            )
            run_dbs_identifier.backtesting_id = await run_dbs_identifier.generate_new_backtesting_id()
            if self.octobot_backtesting.enable_storage:
                # initialize to lock the backtesting id
                await run_dbs_identifier.initialize()
            self.backtesting_config[common_constants.CONFIG_BACKTESTING_ID] = run_dbs_identifier.backtesting_id

    def _find_reference_market_and_update_contract_type(self):
        ref_market_candidate = None
        ref_market_candidates = {}
        forced_contract_type = self.octobot_origin_config.get(common_constants.CONFIG_CONTRACT_TYPE,
                                                              common_constants.USE_CURRENT_PROFILE)
        for symbols in self.symbols_to_create_exchange_classes.values():
            symbol = symbols[0]
            if next(iter(self.octobot_backtesting.exchange_type_by_exchange.values())) \
                    == common_constants.CONFIG_EXCHANGE_FUTURE:
                if forced_contract_type == common_constants.USE_CURRENT_PROFILE:
                    if symbol.is_inverse():
                        if not all([symbol.is_inverse() for symbol in symbols]):
                            self.logger.error(f"Mixed inverse and linear contracts backtesting are not supported yet")
                            self.octobot_backtesting.futures_contract_type = \
                                trading_enums.FutureContractType.INVERSE_PERPETUAL
                    else:
                        if not all([symbol.is_linear() for symbol in symbols]):
                            self.logger.error(f"Mixed inverse and linear contracts backtesting are not supported yet")
                        self.octobot_backtesting.futures_contract_type = \
                            trading_enums.FutureContractType.LINEAR_PERPETUAL
                    # in inverse contracts, use BTC for BTC/USD trading as reference market
                    if symbol.settlement_asset:
                        # only use settlement asset if available
                        return symbol.settlement_asset
                else:
                    self.octobot_backtesting.futures_contract_type = forced_contract_type
                    return symbol.base \
                        if forced_contract_type is trading_enums.FutureContractType.INVERSE_PERPETUAL \
                        else symbol.quote

            for symbol in symbols:
                quote = symbol.quote
                if ref_market_candidate is None:
                    ref_market_candidate = quote
                if quote in ref_market_candidates:
                    ref_market_candidates[quote] += 1
                else:
                    ref_market_candidates[quote] = 1
                if ref_market_candidate != quote and \
                        ref_market_candidates[ref_market_candidate] < ref_market_candidates[quote]:
                    ref_market_candidate = quote
        return ref_market_candidate

    def _add_config_default_backtesting_values(self):
        if backtesting_constants.CONFIG_BACKTESTING not in self.backtesting_config:
            self.backtesting_config[backtesting_constants.CONFIG_BACKTESTING] = {}
        self.backtesting_config[backtesting_constants.CONFIG_BACKTESTING][common_constants.CONFIG_ENABLED_OPTION] = True
        self.backtesting_config[common_constants.CONFIG_TRADER][common_constants.CONFIG_ENABLED_OPTION] = False
        self.backtesting_config[common_constants.CONFIG_SIMULATOR][common_constants.CONFIG_ENABLED_OPTION] = True

    def _add_crypto_currencies_config(self):
        for symbols in self.symbols_to_create_exchange_classes.values():
            for symbol in symbols:
                symbol_id = str(symbol)
                if symbol_id not in self.backtesting_config[common_constants.CONFIG_CRYPTO_CURRENCIES]:
                    self.backtesting_config[common_constants.CONFIG_CRYPTO_CURRENCIES][symbol_id] = {
                        common_constants.CONFIG_CRYPTO_PAIRS: []
                    }
                    self.backtesting_config[common_constants.CONFIG_CRYPTO_CURRENCIES][symbol_id][
                        common_constants.CONFIG_CRYPTO_PAIRS] = [symbol_id]



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import copy

import octobot.api.backtesting as octobot_backtesting_api
import octobot.strategy_optimizer as octobot_strategy_optimizer
import octobot.backtesting as octobot_backtesting
import octobot_commons.constants as commons_constants

import octobot_backtesting.errors as backtesting_errors

import octobot_evaluators.constants as evaluator_constants

import octobot_trading.api as trading_api


class StrategyTestSuite(octobot_backtesting.AbstractBacktestingTest):
    # set to True to skip bigger scenarii and make tests faster
    SKIP_LONG_STEPS = False

    def __init__(self):
        super().__init__()
        self._profitability_results = []
        self._trades_counts = []
        self.current_progress = 0
        self.exceptions = []
        self.evaluators = []

    def get_test_suite_result(self):
        return octobot_strategy_optimizer.TestSuiteResult(self._profitability_results,
                                                          self._trades_counts,
                                                          self.config[commons_constants.CONFIG_TRADING][
                                                              commons_constants.CONFIG_TRADER_RISK],
                                                          self.config[evaluator_constants.CONFIG_FORCED_TIME_FRAME],
                                                          self.evaluators,
                                                          self.strategy_evaluator_class.get_name())

    async def run_test_suite(self, strategy_tester):
        self.exceptions = []
        tests = [self.test_slow_downtrend, self.test_sharp_downtrend, self.test_flat_markets,
                 self.test_slow_uptrend, self.test_sharp_uptrend, self.test_up_then_down]
        print('| ', end='')
        nb_tests = len(tests)
        for i, test in enumerate(tests):
            try:
                await test(strategy_tester)
            except Exception as e:
                self.logger.exception(e, True, f"Exception when running test {test.__name__}: {e}")
                self.exceptions.append(e)
            finally:
                self.current_progress = int((i + 1) / nb_tests * 100)
            print('#', end='')
        print(' |', end='')
        return not self.exceptions

    async def test_default_run(self, strategy_tester):
        await strategy_tester.run_test_default_run(None)

    async def test_slow_downtrend(self, strategy_tester):
        await strategy_tester.run_test_slow_downtrend(None, None, None, None, StrategyTestSuite.SKIP_LONG_STEPS)

    async def test_sharp_downtrend(self, strategy_tester):
        await strategy_tester.run_test_sharp_downtrend(None, None, StrategyTestSuite.SKIP_LONG_STEPS)

    async def test_flat_markets(self, strategy_tester):
        await strategy_tester.run_test_flat_markets(None, None, None, None, StrategyTestSuite.SKIP_LONG_STEPS)

    async def test_slow_uptrend(self, strategy_tester):
        await strategy_tester.run_test_slow_uptrend(None, None)

    async def test_sharp_uptrend(self, strategy_tester):
        await strategy_tester.run_test_sharp_uptrend(None, None)

    async def test_up_then_down(self, strategy_tester):
        await strategy_tester.run_test_up_then_down(None, StrategyTestSuite.SKIP_LONG_STEPS)

    def _handle_results(self, independent_backtesting, profitability):
        trades_count = 0
        profitability_result = None
        skip_this_run = False
        if independent_backtesting is not None:
            exchange_manager_ids = octobot_backtesting_api.get_independent_backtesting_exchange_manager_ids(
                independent_backtesting)
            try:
                for exchange_manager in trading_api.get_exchange_managers_from_exchange_ids(exchange_manager_ids):
                    _, profitability, _, market_average_profitability, _ = \
                        trading_api.get_profitability_stats(exchange_manager)
                    # Only one exchange manager per run
                    profitability_result = (profitability, market_average_profitability)
                    trades_count += len(trading_api.get_trade_history(exchange_manager))
            except (AttributeError, KeyError):
                skip_this_run = True
            if not skip_this_run:
                if profitability_result is None:
                    raise RuntimeError("Error with independent backtesting: no available exchange manager")
                self._profitability_results.append(profitability_result)
                self._trades_counts.append(trades_count)

    async def _run_backtesting_with_current_config(self, data_file_to_use):
        independent_backtesting = None
        try:
            config_to_use = copy.deepcopy(self.config)
            independent_backtesting = octobot_backtesting_api.create_independent_backtesting(
                config_to_use,
                self.tentacles_setup_config,
                [data_file_to_use],
                "",
                enforce_total_databases_max_size_after_run=False,
                enable_storage=False,
            )
            await octobot_backtesting_api.initialize_and_run_independent_backtesting(independent_backtesting, log_errors=False)
            await octobot_backtesting_api.join_independent_backtesting(independent_backtesting)
            return independent_backtesting
        except backtesting_errors.MissingTimeFrame:
            # ignore this exception: is due to missing of the only required time frame
            return independent_backtesting
        except Exception as e:
            self.logger.exception(e, True, str(e))
            return independent_backtesting



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import pytest
import mock

import octobot.automation
import octobot_tentacles_manager.api as tentacles_manager_api
import octobot_commons.asyncio_tools as asyncio_tools
import tests.test_utils.config as config
import tests.functional_tests.automations as test_automations


@pytest.fixture
def automation():
    import tentacles
    tentacles_manager_api.reload_tentacle_info()
    return octobot.automation.Automation("bot_id", config.load_test_tentacles_config())


@pytest.mark.asyncio
async def test_empty_initialize(automation):
    await automation.initialize()
    assert automation.automation_details == []


def test_get_all_steps(automation):
    # ensure tentacles automations are available (counts are all + 1 due to testing steps)
    all_events, all_conditions, all_actions = automation.get_all_steps()
    assert len(all_events) > 2
    assert len(all_conditions) > 1
    assert len(all_actions) > 2


@pytest.mark.asyncio
async def test_automation_workflow(automation):
    test_config = {
        automation.AUTOMATIONS_COUNT: 1,
        automation.AUTOMATIONS: {
            "1": {
                automation.TRIGGER_EVENT: test_automations.TestTriggerEvent.get_name(),
                automation.CONDITIONS: [test_automations.TestCondition.get_name()],
                automation.ACTIONS: [test_automations.TestAction.get_name()],
                test_automations.TestTriggerEvent.get_name(): {},
                test_automations.TestCondition.get_name(): {},
                test_automations.TestAction.get_name(): {},
            }
        }
    }
    with mock.patch.object(tentacles_manager_api, "get_tentacle_config", mock.Mock(return_value=test_config)) \
         as get_tentacle_config_mock:
        await automation.initialize()
        get_tentacle_config_mock.assert_called_once()

        assert len(automation.automation_details) == 1
        automation_detail = automation.automation_details[0]
        assert isinstance(automation_detail.trigger_event, test_automations.TestTriggerEvent)
        assert len(automation_detail.conditions) == 1
        assert len(automation_detail.actions) == 1
        assert isinstance(automation_detail.conditions[0], test_automations.TestCondition)
        assert isinstance(automation_detail.actions[0], test_automations.TestAction)

        automation_detail.trigger_event.get_next_event_mock.assert_not_awaited()
        automation_detail.conditions[0].evaluate_mock.assert_not_awaited()
        automation_detail.actions[0].process_mock.assert_not_awaited()

        # wait for automation task to process
        await asyncio_tools.wait_asyncio_next_cycle()

        # ensure automation went through trigger event, condition and action
        automation_detail.trigger_event.get_next_event_mock.assert_awaited_once()
        automation_detail.conditions[0].evaluate_mock.assert_awaited_once()
        automation_detail.actions[0].process_mock.assert_awaited_once()



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.tentacles_management.class_inspector as class_inspector
import octobot.strategy_optimizer.strategy_design_optimizer as strategy_design_optimizer


def create_most_advanced_strategy_design_optimizer(
        trading_mode, config, tentacles_setup_config, optimizer_settings=None
):
    advanced_class = strategy_design_optimizer.StrategyDesignOptimizer
    optimizer_classes = [
        optimizer_class
        for optimizer_class in class_inspector.get_all_classes_from_parent(advanced_class)
        if optimizer_class.ALLOWED_IN_FACTORY
    ]
    if optimizer_classes:
        # the last one of the list is the most advanced one
        advanced_class = optimizer_classes[-1]
    return advanced_class(trading_mode, config, tentacles_setup_config, optimizer_settings=optimizer_settings)



#  This file is part of OctoBot (https://github.com/Drakkar-Software/OctoBot)
#  Copyright (c) 2023 Drakkar-Software, All rights reserved.
#
#  OctoBot is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  OctoBot is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with OctoBot. If not, see <https://www.gnu.org/licenses/>.
import enum


class CommunityFeedType(enum.Enum):
    WebsocketFeed = "WebsocketFeed"
    MQTTFeed = "MQTTFeed"
    SupabaseFeed = "SupabaseFeed"


class CommunityEnvironments(enum.Enum):
    Staging = "Staging"
    Production = "Production"


class OptimizerModes(enum.Enum):
    NORMAL = "normal"
    GENETIC = "genetic"


class OptimizerConfig(enum.Enum):
    OPTIMIZER_ID = "optimizer_id"
    OPTIMIZER_IDS = "optimizer_ids"
    RANDOMLY_CHOSE_RUNS = "randomly_chose_runs"
    DATA_FILES = "data_files"
    OPTIMIZER_CONFIG = "optimizer_config"
    EXCHANGE_TYPE = "exchange_type"
    QUEUE_SIZE = "queue_size"
    EMPTY_THE_QUEUE = "empty_the_queue"
    START_TIMESTAMP = "start_timestamp"
    END_TIMESTAMP = "end_timestamp"
    IDLE_CORES = "idle_cores"
    NOTIFY_WHEN_COMPLETE = "notify_when_complete"
    DB_UPDATE_PERIOD = "db_update_period"
    MODE = "mode"
    MAX_OPTIMIZER_RUNS = "max_optimizer_runs"
    INITIAL_GENERATION_COUNT = "initial_generation_count"
    DEFAULT_GENERATIONS_COUNT = "default_generations_count"
    DEFAULT_RUN_PER_GENERATION = "default_run_per_generation"
    DEFAULT_SCORING_PARAMETERS = "default_scoring_parameters"
    DEFAULT_OPTIMIZER_FILTERS = "default_optimizer_filters"
    DEFAULT_OPTIMIZER_CONSTRAINTS = "default_optimizer_constraints"
    DEFAULT_MUTATION_PERCENT = "default_mutation_percent"
    MAX_MUTATION_PROBABILITY_PERCENT = "max_mutation_probability_percent"
    MIN_MUTATION_PROBABILITY_PERCENT = "min_mutation_probability_percent"
    DEFAULT_MAX_MUTATION_NUMBER_MULTIPLIER = "default_max_mutation_number_multiplier"
    DEFAULT_CROSSOVER_PERCENT = "default_crossover_percent"
    STAY_WITHIN_BOUNDARIES = "stay_within_boundaries"
    TARGET_FITNESS_SCORE = "target_fitness_score"
